import logging
import argparse
import sys
import re
import json

import numpy as np

from triplix.core import concatemers
from triplix.core import tri_alignment, triplet
from triplix.core import configurations
from triplix.core import stat
from triplix.core import plot

logger = logging.getLogger(__name__)


def cli_concatemers(cli_args):
    exporter = concatemers.ExporterHDF5(
        bam_path=cli_args.input,
        assembly=cli_args.assembly,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        assume_sorted=cli_args.assume_sorted,
        cell_type=cli_args.cell_type,
        assay_name=cli_args.assay_name,
        experiment_name=cli_args.experiment_name,
    )
    exporter.initialize_hdf5()
    exporter.store_concatemers()


def cli_tri_alignments(cli_args):
    extractor = tri_alignment.extract.TriAlignmentExtractor(
        concatemers_path=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
    )
    extractor.extract_tri_alignments()


def cli_sort(cli_args):
    sorter = tri_alignment.sort.TriAlignmentSorter(
        trialign_path=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
    )
    sorter.sort_tri_alignments(n_thread=cli_args.threads,
                               memory=cli_args.memory)


def cli_merge(cli_args):
    if cli_args.metadata is None:
        metadata = {}
    else:
        metadata = dict(param.split(':') for param in cli_args.metadata.split(','))

    merger = tri_alignment.merge.TripletMerger(
        input_patterns=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        concatenate_only=cli_args.concatenate,
        check_coverage=cli_args.check_coverage,
        sort_by=cli_args.sort_by,
        metadata=metadata,
    )
    merger.merge_triplets(n_thread=cli_args.threads,
                          memory=cli_args.memory,
                          ignore_headers=cli_args.ignore_samheaders)


def cli_index(cli_args):
    indexer = tri_alignment.index.TripletIndexer(
        triplet_path=cli_args.input,
    )
    indexer.index()


def cli_bin(cli_args):
    binner = triplet.bin.TriAlignmentBinner(
        input_path=cli_args.input,
        anchor_width=int(cli_args.anchor_width),
        anchor_max_distance=int(cli_args.anchor_max_distance),
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
    )
    binner.bin(
        chunk_chrom=cli_args.chunk_chrom,
        chunk_start=cli_args.chunk_start,
        chunk_end=cli_args.chunk_end,
        mapping_quality=cli_args.mapping_quality,
    )


def cli_transform(cli_args):
    # process transform's parameters
    params = json.loads(cli_args.transform_params)

    if cli_args.transform_name == 'Gaussian':

        smoother = triplet.transform.GaussianSmoother(
            triplet_path=cli_args.input,
        )
        smoother.smooth(
            chunk_chrom=cli_args.chunk_chrom,
            chunk_start=cli_args.chunk_start,
            chunk_end=cli_args.chunk_end,
            kernel_width=params['kernel_width'],
            kernel_scale=params['kernel_scale'],
            min_distance=int(cli_args.min_distance),
            max_distance=int(cli_args.max_distance),
            label=cli_args.transform_label,
        )
    else:
        raise ValueError(f'Unknown transformation: {cli_args.transform_name}')


def cli_kdtree(cli_args):
    chroms = cli_args.chroms
    if chroms is not None:
        chroms = chroms.split(',')

    nei_finder = triplet.kdtree.NeighborFinder(
        triplet_path=cli_args.input,
    )
    nei_finder.find_neighbors(
        factors_info=cli_args.factors,
        target=cli_args.target,
        chroms=chroms,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        n_samples=cli_args.n_samples,
    )


def cli_enrichment(cli_args):
    enrich_estimator = triplet.enrichment.EnrichmentEstimator(
        triplet_path=cli_args.input,
        kdtree_path=cli_args.kdtree,
    )
    enrich_estimator.estimate_enrichments(
        chunk_chrom=cli_args.chunk_chrom,
        chunk_start=cli_args.chunk_start,
        chunk_end=cli_args.chunk_end,
        n_neighbors=cli_args.n_neighbors,
        view_point=cli_args.view_point,
    )


def cli_plot_virtual_hic(cli_args):
    # process plot params
    if cli_args.plot_params is None:
        plot_params = {}
    else:
        plot_params = json.loads(cli_args.plot_params)

    # plot_params = {}
    # if cli_args.params is not None:
    #     for param_str in cli_args.params.split('|'):
    #         col_name, param_name, param_value = re.split(r'[:=]', param_str)
    #         if col_name not in plot_params:
    #             plot_params[col_name] = {}
    #         plot_params[col_name][param_name] = param_value

    # if param_name in ['clim_range']:
    #     plot_params[col_name][param_name] = [float(value) for value in param_value.split(',')]
    #     plot_params[col_name][param_name] = param_value
    # else:
    #     plot_params[col_name][param_name] = param_value

    plot.plot_virtual_hic(
        input_patterns=cli_args.input,
        view_chrom=cli_args.view_chrom,
        view_start=cli_args.view_start,
        view_end=cli_args.view_end,
        view_point=cli_args.view_point,
        anchor_width=cli_args.anchor_width,
        plot_names=cli_args.columns.split(','),
        mapping_quality=cli_args.mapping_quality,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        plot_params=plot_params,
    )


def cli_plot_hic(cli_args):
    # process plot params
    plot_params = {}
    if cli_args.params is not None:
        for param_str in cli_args.params.split('|'):
            col_name, param_name, param_value = re.split(r'[:=]', param_str)
            if col_name not in plot_params:
                plot_params[col_name] = {}
            if param_name in ['clim_range']:
                plot_params[col_name][param_name] = [float(value) for value in param_value.split(',')]
            else:
                plot_params[col_name][param_name] = param_value

    plot.plot_hic(
        input_patterns=cli_args.input,
        view_chrom=cli_args.view_chrom,
        view_start=cli_args.view_start,
        view_end=cli_args.view_end,
        anchor_width=cli_args.anchor_width,
        plot_names=cli_args.columns.split(','),
        mapping_quality=cli_args.mapping_quality,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        plot_params=plot_params,
    )


def cli_plot_processing_history(cli_args):
    # set igraph parameters
    kwargs = {}
    if cli_args.igraph_params is not None:
        for param in cli_args.igraph_params.split(','):
            key, value = param.split(':')
            if key == 'figsize':
                kwargs[key] = [float(field) for field in value.split('-')]
            else:
                kwargs[key] = value
    kwargs['title'] = kwargs.get('title', f'{cli_args.input}')

    plot.plot_processing_history(
        input_path=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        **kwargs,
    )


def cli_plot_read_length_histogram(cli_args):
    plot.plot_read_length_histogram(
        concatemers_path=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
    )


def cli_plot_alignment_length_histogram(cli_args):
    plot.plot_alignment_length_histogram(
        concatemers_path=cli_args.input,
        output_dir=cli_args.output_dir,
        output_name=cli_args.output_name,
        mapping_quality=cli_args.mapping_quality,
        max_length=cli_args.max_length,
    )


def cli_stat_decay(cli_args):
    if cli_args.analysis == 'decay':
        stat.collect_contact_decay_stats(
            concatemer_path=cli_args.input,
            output_dir=cli_args.output_dir,
            output_name=cli_args.output_name,
        )
    else:
        raise ValueError(f'Unknown stat: {cli_args.analysis}')


def parse_cli_arguments():
    """Generates command line arguments
    """
    parser_main = argparse.ArgumentParser(prog='triplix', add_help=False)
    commands = parser_main.add_subparsers(
        title='required commands are',
        dest='command',
        # help='Description:'
    )
    parser_main.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS, help='Show this help message and exit.')
    parser_main.add_argument('--debug', action='store_true', help='Activate the debug mode.')
    parser_main.add_argument('--log_level', type=int, default=logging.INFO, help='Log level. An integer in [0, 50] interval. With larger values Triplix will be more conservative in printing messages (default: %(default)s)')

    #########################################
    # forming concatemers
    parser_concatemers = commands.add_parser('concatemers', help='Parse input BAM file and extract Concatemers')
    optional = parser_concatemers._action_groups.pop()
    required = parser_concatemers.add_argument_group('required arguments')
    required.add_argument('--input', required=True, metavar='<PATH>', help='Path to the input BAM file. The BAM file should be sorted by `read name` (i.e., using `samtools sort -n`)')
    optional.add_argument('--assembly', default='unknown', metavar='<STRING>', help='Name of the assembly (or genome) such as `hg19`, `hg38`, `mm10`, etc. (default: %(default)s)')
    optional.add_argument('--output_dir', default='./concatemers/', metavar='<PATH>',
                          help='Output directory where the output Concatemers file (.concatemers.h5) will be stored (default: %(default)s)')
    optional.add_argument('--output_name', default=None, metavar='<FILENAME>',
                          help='Output Concatemers file name. If not provided, the `experiment_name` is used with an added `.concatemers.h5` extension')
    optional.add_argument('--assume_sorted', action='store_true', help='Disables the name-sort check and assumes that the input BAM file is name sorted (default: %(default)s)')
    optional.add_argument('--cell_type', default='unknown', metavar='<STRING>', help='Name of the cell-type that is used to produce the current experiment (default: %(default)s)')
    optional.add_argument('--assay_name', default='unknown', metavar='<STRING>', help='Name of the assay that is used to produce the current experiment (default: %(default)s)')
    optional.add_argument('--experiment_name', default=None, metavar='<STRING>', help='A name for the current experiment. If not provided, the input BAM filename will be used as the `--experiment_name` instead.')
    parser_concatemers._action_groups.append(optional)
    parser_concatemers.set_defaults(func=cli_concatemers)

    #########################################
    # extracting tri-alignments
    parser_tri_alignments = commands.add_parser('tri_alignments', help='Process a Concatemers file (.concatemers.h5) and extract Tri-alignments from it')
    parser_tri_alignments.add_argument('--input', required=True, metavar='<PATH>', help='Path to input Concatemers file (e.g. `*.concatemers.h5`)')
    parser_tri_alignments.add_argument('--output_dir', default='./tri-alignments/', metavar='<PATH>',
                                       help='Output directory where the extracted Tri-alignments will be stored (default: %(default)s)')
    parser_tri_alignments.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                       help='Output file name for the extracted Tri-alignments. If not provided, the output is stored using the input Concatemers file name, with a `.tri-alignments.tsv.bgz` extension')
    parser_tri_alignments.set_defaults(func=cli_tri_alignments)

    #########################################
    # sorting tri-alignments
    parser_sort = commands.add_parser('sort', help='Sort a Tri-alignments file according to the chromosomal positions of its Tri-alignments anchors.')
    parser_sort.add_argument('--input', required=True, metavar='<PATH>', help='Path to Tri-alignments file (e.g. `*.tri-alignments.tsv.bgz`)')
    parser_sort.add_argument('--output_dir', default='./tri-alignments_sorted', metavar='<PATH>',
                             help='Output directory where the sorted Tri-alignments file will be stored (default: %(default)s)')
    parser_sort.add_argument('--output_name', default=None, metavar='<FILENAME>',
                             help='Output name for the sorted Tri-alignments file. If not provided, the output is stored using the input Tri-alignments filename, with a `.sorted.tri-alignments.tsv.bgz` extension')
    parser_sort.add_argument('--memory', default='2G', metavar='<FLOAT>G', help='Amount of memory (in Gigabyte, or another supported size unit) that can be used by `sort` program (default: "%(default)s")')
    parser_sort.add_argument('--threads', default=1, type=int, metavar='<INTEGER>', help='Number of threads that can be used by `sort` program (default: "%(default)s")')
    parser_sort.set_defaults(func=cli_sort)

    #########################################
    # merging tri-alignments
    parser_merge = commands.add_parser('merge', help='Merge a series of Tri-alignments files')
    parser_merge.add_argument('--input', required=True, metavar='<PATH>[,<PATH>]', help='A comma-separated list of paths to Tri-alignments files (e.g. `*.tri-alignments.tsv.bgz`). Each PATH can be a pattern that contain `*` to match several files.')
    parser_merge.add_argument('--output_dir', default='./tri-alignments_merged/', metavar='<PATH>',
                              help='Output directory where the merged triplet will be stored (default: %(default)s)')
    parser_merge.add_argument('--output_name', default=None, metavar='<FILENAME>',
                              help='Output file name for the merged Tri-alignments. If not provided, the output is stored using the first input Tri-alignments filename, with a `.merged.tri-alignments.tsv.bgz` extension')
    parser_merge.add_argument('--memory', default='2G', help='Amount of memory (in Gigabyte unit) that can be used by `merge` program (default: "%(default)s")')
    parser_merge.add_argument('--threads', default=4, type=int, metavar='<INTEGER>', help='Number of threads that can be used by `sort` program (default: "%(default)s")')
    parser_merge.add_argument('--concatenate', action="store_true", help='Perform a simple concatenation of Tri-alignments files instead of merging files in a sorted manner')
    parser_merge.add_argument('--check_coverage', action="store_true", help='Checks whether merging chunks are overlapping, or if a chromosomal region is not covered by any chunk')
    parser_merge.add_argument('--sort_by', choices=['none', 'chrom'], default='chrom',
                              help='Sort the input files before merge/concatenation. `none` ignores sorting. `chrom` sorts files by the order of chromosomes in the original BAM file (default: "%(default)s")')
    # todo: add possibility to update the header/meta-data of the merged dataset, e.g. experiment_name
    # todo: implement a better header merger. Then remove this flag
    parser_merge.add_argument('--ignore_samheaders', action="store_true", help='Ignores merging sam headers.')
    parser_merge.add_argument('--metadata', default=None, metavar='<STRING>:<STRING>,[...]', help='A comma-separated list of key:values to be stored in the Tri-alignments file header such as experiment or assay name')
    parser_merge.set_defaults(func=cli_merge)

    #########################################
    # indexing tri-alignments
    parser_index = commands.add_parser('index', help='Index a tri-alignment file')
    parser_index.add_argument('--input', required=True, metavar='<PATH>', help='Path to a Tri-alignments file (e.g. `*.tri-alignments.tsv.bgz`). The index will be stored besides the input Tri-alignments file')
    parser_index.set_defaults(func=cli_index)

    #########################################
    # mapping tri-alignments into bins
    parser_bin = commands.add_parser('bin', help='Discretize the genome into equally-spaced anchors (i.e., bins) and represent the interactions in terms of Triplets')
    parser_bin.add_argument('--input', required=True, metavar='<PATH>', help='Path to input Tri-alignments file (e.g. `*.tri-alignments.tsv.bgz`).')
    parser_bin.add_argument('--output_dir', default='./triplets/', metavar='<PATH>',
                            help='Output directory where the Triplets file will be stored (default: %(default)s)')
    parser_bin.add_argument('--output_name', default=None, metavar='<FILENAME>',
                            help='Output name for the Triplets file. If not provided, the output is stored using the input Tri-alignments file name, with a `.triplets.h5` extension')
    parser_bin.add_argument('--chunk_chrom', required=True, metavar='<STRING>', help='Name of the chromosome from which triplet are formed')
    parser_bin.add_argument('--chunk_start', required=True, type=float, metavar='<INTEGER>', help='Start position of the chunk from which triplets are formed')
    parser_bin.add_argument('--chunk_end', required=True, type=float, metavar='<INTEGER>', help='End position of the chunk from which triplets are formed')
    parser_bin.add_argument('--anchor_width', default=int(25e3), type=float, metavar='<INTEGER>', help='Width of each anchor (i.e., genomic bin) (default: "%(default)s")')
    parser_bin.add_argument('--anchor_max_distance', default=int(2e6), type=float, metavar='<INTEGER>', help='Maximum distance between anchors of a triplet (default: "%(default)s")')
    parser_bin.add_argument('--mapping_quality', default=5, type=int, metavar='<INTEGER>', help='Minimum mapping quality threshold to consider an alignment to be valid (default: "%(default)s")')
    parser_bin.set_defaults(func=cli_bin)

    #########################################
    # transform triplet counts
    parser_transform = commands.add_parser('transform', help='Transform the raw counts using the requested transformation.')
    parser_transform.add_argument('--input', required=True, metavar='<PATH>', help='Path to Triplets file (e.g. `*.triplets.h5`). This file needs to be indexed.')
    parser_transform.add_argument('--transform_name', choices=['Gaussian'], default='Gaussian', help='Transformation name')
    parser_transform.add_argument('--transform_params', default='{"kernel_width":15,"kernel_scale":1.0}', help='JSON-formatted parameters of the transformation (default: "%(default)s")')
    parser_transform.add_argument('--transform_label', default='observed', metavar='<STRING>', help='Each transformed column will be labeled using this parameter (default: "%(default)s")')
    parser_transform.add_argument('--chunk_chrom', required=True, metavar='<STRING>', help='Name of the chromosome from which triplets are collected')
    parser_transform.add_argument('--chunk_start', required=True, type=float, metavar='<INTEGER>', help='Start position of the chunk from which triplets are collected')
    parser_transform.add_argument('--chunk_end', required=True, type=float, metavar='<INTEGER>', help='End position of the chunk from which triplets are collected')
    parser_transform.add_argument('--min_distance', default=int(25e3), type=float, metavar='<INTEGER>', help='Minimum distance between anchors of a triplet (default: "%(default)s")')
    parser_transform.add_argument('--max_distance', default=int(2e6), type=float, metavar='<INTEGER>', help='Maximum distance between anchors of a triplet (default: "%(default)s")')
    parser_transform.set_defaults(func=cli_transform)

    #########################################
    # finding neighbors
    parser_findnei = commands.add_parser('kdtree', help='Construct a KDTree structure to efficiently find Triplet neighbors.')
    parser_findnei.add_argument('--input', required=True, metavar='<PATH>', help='Path to Triplets file (e.g. `*.triplets.h5`).')
    parser_findnei.add_argument('--output_dir', default='./kdtrees/', metavar='<PATH>',
                                help='Output directory where the KDTree will be stored (default: %(default)s)')
    parser_findnei.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                help='Output file name for the KDTree. If not provided, the output is stored using the input Triplets filename, with a `.kdtree.joblib` extension.')
    parser_findnei.add_argument('--chroms', metavar='STRING[,STRING]', default=None, help='A comma-separated list of chromosome names to be used in KDTree construction. If not provided, all chromosomes will be used.')
    parser_findnei.add_argument('--factors', required=True, metavar='<PATH>|<COL>:<FUNC>[,<COL>:<FUNC>]', help='Path to correction factor definitions, or a comma-separated list of <column_name>:<normalization>')
    parser_findnei.add_argument('--target', required=True, default='observed_ABC', metavar='<STRING>', help='Name of the column to be used as "target", in the background model  (default: "%(default)s")')
    parser_findnei.add_argument('--n_samples', default=np.inf, type=float, help='Sets the maximum number of samples to be included in the KDTree (default: "%(default)s")')
    parser_findnei.set_defaults(func=cli_kdtree)

    #########################################
    # estimate enrichments
    parser_enrich = commands.add_parser('enrichment', help='Estimate the enrichments of (a chunk of) triplet.')
    parser_enrich.add_argument('--input', required=True, metavar='<PATH>', help='Path to Triplets file (e.g. `*.triplets.h5`).')
    parser_enrich.add_argument('--kdtree', required=True, metavar='<PATH>', help='Path to KDTree, constructed from the input Triplets file.')
    parser_enrich.add_argument('--chunk_chrom', required=True, metavar='<STRING>', help='Name of the chromosome from which the triplets are going to be collected')
    parser_enrich.add_argument('--chunk_start', required=True, type=float, metavar='<INTEGER>', help='Start position of the chunk from which the triplets are going to be collected')
    parser_enrich.add_argument('--chunk_end', required=True, type=float, metavar='<INTEGER>', help='End position of the chunk from which the triplets are going to be collected')
    parser_enrich.add_argument('--view_point', default=None, type=float, metavar='<INTEGER>', help='If given, only triplets that overlap with the viewpoint are considered')
    parser_enrich.add_argument('--n_neighbors', default=int(10e3), type=int, metavar='<INTEGER>', help='Number of neighbors to use during enrichment estimation (default: "%(default)s")')
    parser_enrich.set_defaults(func=cli_enrichment)

    #########################################
    # plotting
    parser_plot = commands.add_parser('plot', help='Plotting varied characteristic of a given input file.')
    command_plot = parser_plot.add_subparsers(
        title='available plots are',
        dest='type',
    )

    # plotting virtual-hic
    parser_virthic = command_plot.add_parser('virtual_hic', help='Generate a virtual HiC plot, depicting the requested columns (using `--columns`).')
    parser_virthic.add_argument('--input', required=True, metavar='<PATH>',
                                help='Path to input file. Multiple paths could be provided using comma-separated format. '
                                     + 'Wild chars (i.e., "*") are also supported. The input file could be a BAM, tri-alignment, or triplets file. '
                                     + 'In case of BAM or tri-alignment, the file  needs to be indexed.')
    parser_virthic.add_argument('--output_dir', default='./plots/', metavar='<PATH>',
                                help='Output directory where the plot will be stored (default: %(default)s)')
    parser_virthic.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                help='Output file name for the plot. If not provided, the output is stored using the input filename, with a `.pdf` extension.')
    parser_virthic.add_argument('--view_chrom', required=True, metavar='<STRING>', help='Name of the chromosome from which contact data are collected')
    parser_virthic.add_argument('--view_start', required=True, type=float, metavar='<INTEGER>', help='Start position of the chunk from which counts are collected')
    parser_virthic.add_argument('--view_end', required=True, type=float, metavar='<INTEGER>', help='End position of the chunk from which contact data are collected')
    parser_virthic.add_argument('--view_point', required=True, type=float, metavar='<INTEGER>', help='View point position that all contact data will interact to')
    parser_virthic.add_argument('--anchor_width', default=None, type=float, metavar='<INTEGER>', help='Width of each anchor (i.e., genomic bin). Only when then input file is a tri-alignment (default: "%(default)s")')
    parser_virthic.add_argument('--columns', default='count_AB,count_AC,count_ABC', metavar='<STRING>', help='A comma-separated list of column names that should be plotted (default: "%(default)s")')
    parser_virthic.add_argument('--plot_params', default=None, metavar='<STRING>', help='JSON-formatted parameters of each plot (default: "%(default)s")')
    parser_virthic.add_argument('--mapping_quality', default=5, type=int, metavar='<INTEGER>', help='Minimum mapping quality threshold to consider an alignment to be valid (default: "%(default)s")')
    parser_virthic.set_defaults(func=cli_plot_virtual_hic)

    # plotting hic
    parser_hic = command_plot.add_parser('hic', help='Generate a pairwise HiC plot, depicting the requested columns (using `--columns`).')
    parser_hic.add_argument('--input', required=True, metavar='<PATH>', help='Path to input BAM files. Multiple paths could be provided using comma-separated format. Wild chars (i.e., "*") are also supported.')
    parser_hic.add_argument('--output_dir', default='./plots/', metavar='<PATH>',
                            help='Output directory where the plot will be stored (default: %(default)s)')
    parser_hic.add_argument('--output_name', default=None, metavar='<FILENAME>',
                            help='Output file name for the plot. If not provided, the output is stored using the input filename, with a `.pdf` extension.')
    parser_hic.add_argument('--view_chrom', required=True, metavar='<STRING>', help='Name of the chromosome from which contact data are collected')
    parser_hic.add_argument('--view_start', required=True, type=float, metavar='<INTEGER>', help='Start position of the chunk from which counts are collected')
    parser_hic.add_argument('--view_end', required=True, type=float, metavar='<INTEGER>', help='End position of the chunk from which contact data are collected')
    parser_hic.add_argument('--anchor_width', default=None, type=float, metavar='<INTEGER>', help='Width of each anchor (i.e., genomic bin). Only when then input file is a tri-alignment (default: "%(default)s")')
    parser_hic.add_argument('--columns', default='count_AB,count_AC,count_ABC', metavar='<STRING>', help='A comma-separated list of column names that should be plotted (default: "%(default)s")')
    parser_hic.add_argument('--params', default=None, metavar='COLUMN-NAME:PARAM-NAME=PARAM-VALUE[,...]', help='A comma-separated list of column names and their corresponding parameteres and values')
    parser_hic.add_argument('--mapping_quality', default=5, type=int, metavar='<INTEGER>', help='Minimum mapping quality threshold to consider an alignment to be valid (default: "%(default)s")')
    parser_hic.set_defaults(func=cli_plot_hic)

    # plotting processing history
    parser_proc_hist = command_plot.add_parser('processing_history', help='Plot the processing history of the given input file.')
    parser_proc_hist.add_argument('--input', required=True, metavar='<PATH>', help='Path to Concatemers, Tri-alignments or Triplets file.')
    parser_proc_hist.add_argument('--output_dir', default='./plots/', metavar='<PATH>',
                                  help='Output directory where the plot will be stored (default: %(default)s)')
    parser_proc_hist.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                  help='Output file name for the plot. If not provided, the output is stored using the input filename, with `.processing-history.pdf` suffix.')
    parser_proc_hist.add_argument('--igraph_params', default=None, metavar='<STRING>:<STRING>[,<STRING>:<STRING>]', help='A comma-separated list of key:value parameters. Will be sent directly to igraph for plotting.')
    parser_proc_hist.set_defaults(func=cli_plot_processing_history)

    # plotting read length histogram
    parser_hist_read_len = command_plot.add_parser('read_length', help='Plot the read length histogram of concatemers stored in the input file.')
    parser_hist_read_len.add_argument('--input', required=True, metavar='<PATH>', help='Path to Concatemers file.')
    parser_hist_read_len.add_argument('--output_dir', default='./plots/', metavar='<PATH>',
                                      help='Output directory where the plot will be stored (default: %(default)s)')
    parser_hist_read_len.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                      help='Output file name for the plot. If not provided, the output is stored using the input filename, following `read-length.<filename>.pdf` pattern.')
    parser_hist_read_len.set_defaults(func=cli_plot_read_length_histogram)

    # plotting alignment length histogram
    parser_hist_aln_len = command_plot.add_parser('alignment_length', help='Plot the alignment length histogram of concatemers stored in the input file.')
    parser_hist_aln_len.add_argument('--input', required=True, metavar='<PATH>', help='Path to Concatemers file.')
    parser_hist_aln_len.add_argument('--output_dir', default='./plots/', metavar='<PATH>',
                                     help='Output directory where the plot will be stored (default: %(default)s)')
    parser_hist_aln_len.add_argument('--output_name', default=None, metavar='<FILENAME>',
                                     help='Output file name for the plot. If not provided, the output is stored using the input filename, following `alignment-length.<filename>.pdf` pattern.')
    parser_hist_aln_len.add_argument('--mapping_quality', default=5, type=int, metavar='<INTEGER>', help='Minimum mapping quality threshold to consider an alignment to be valid (default: "%(default)s")')
    parser_hist_aln_len.add_argument('--max_length', default=int(3e3), type=float, metavar='<INTEGER>', help='Minimum mapping quality threshold to consider an alignment to be valid (default: "%(default)s")')
    parser_hist_aln_len.set_defaults(func=cli_plot_alignment_length_histogram)

    #########################################
    # calculating statistics
    parser_stats = commands.add_parser('stat', help='Generate statistics from a given input file.')
    command_stat = parser_stats.add_subparsers(
        title='available analyses are',
        dest='analysis',
    )

    parser_decay = command_stat.add_parser('decay', help='Calculates decay stats for a given Concatemers file.')
    parser_decay.add_argument('--input', required=True, metavar='<PATH>', help='Path to input Concatemers file (e.g. `*.concatemers.h5`).')
    parser_decay.add_argument('--output_dir', default='./stats/', metavar='<PATH>',
                              help='Output directory where the stats will be stored (default: %(default)s)')
    parser_decay.add_argument('--output_name', default=None, metavar='<FILENAME>',
                              help='Output file name. If not provided, the output is stored using the input Concatemers filename, with a `.tsv` extension.')
    parser_decay.set_defaults(func=cli_stat_decay)

    # processing CLI input arguments
    cli_args = parser_main.parse_args()

    # set global logger
    logger_global = logging.getLogger()
    logger_global.setLevel(cli_args.log_level)

    # activate debug mode, if requested
    configurations.configs['debug'] = cli_args.debug
    if configurations.configs['debug']:
        logger_global.setLevel(logging.DEBUG)
        logger.debug('Debug mode is activated.')

    if cli_args.command is None:
        logger.error('No command is specified. See below for available commands.')
        parser_main.print_help(file=sys.stderr)
        exit(1)
    cli_args.func(cli_args)

    return cli_args



import io
import gzip
import time
import pathlib
import logging
import subprocess
import copy
import json
from collections import deque

import pypairix
import numpy as np
import pandas as pd
import scipy.ndimage as ndimage
import scipy.signal as signal

from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.triplets import TripletsContainer, Triplets
from triplix.core.tri_alignments import TriAlignmentsContainer
from triplix.core.utilities import generate_prog_id, triplet_to_cube, merge_intervals
from triplix.core.hdf5 import HDF5Container

COMMAND_NAME = 'triplix.transform'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


def gauss_kernel_3d(width, sigma):
    if sigma == 0:
        kernel = np.zeros([width, width, width])
        kernel[width // 2, width // 2, width // 2] = 1.0
    else:
        dx, dy, dz = np.mgrid[
                     -width // 2 + 1:width // 2 + 1,
                     -width // 2 + 1:width // 2 + 1,
                     -width // 2 + 1:width // 2 + 1]

        kernel = np.exp(-((dx ** 2 + dy ** 2 + dz ** 2) / (2.0 * sigma ** 2)))
    return kernel / float(kernel.sum())

    # import matplotlib.pyplot as plt
    # from mpl_toolkits.mplot3d import Axes3D
    # krn = (gauss_kernel_3d(10, 2) * 1000).astype(int)
    # z, x, y = krn.nonzero()
    # fig = plt.figure()
    # ax = Axes3D(fig)
    # plt.close('all')
    # fig = plt.figure()
    # ax = fig.gca(projection='3d')
    # ax.scatter(x, y, z, c=-krn[x, y, z], alpha=0.5)
    # plt.show()


class GaussianSmoother:

    def __init__(self, triplet_path):
        self.triplet_path = pathlib.Path(triplet_path).expanduser()

    def smooth(
            self,
            chunk_chrom, chunk_start, chunk_end,
            kernel_width, kernel_scale,
            min_distance, max_distance,
            label
    ):

        # collecting relevant triplets
        with TripletsContainer(container_path=self.triplet_path, exclusive=True) as triplets_obj:
            logger.debug(f'Container is locked, ready to collect triplets: {self.triplet_path}')

            # define the chunk area
            chunk_start = int(chunk_start)
            chunk_end = int(chunk_end)
            anchor_width = triplets_obj.properties['anchor_width']
            chrom_lengths = dict(zip(triplets_obj.properties['chrom_names'], triplets_obj.properties['chrom_lengths']))
            chrom_size = chrom_lengths[chunk_chrom]
            if chunk_start > chrom_size:
                logger.info(f'The requested window ({chunk_start / 1e6:0.2f}-{chunk_end / 1e6:0.2f}mb) '
                            f'is outside of {chunk_chrom} (length={chrom_size / 1e6:0.2f}mb). '
                            f'No processing is needed.')
                return
            logger.info(f'Area of interest is: {chunk_chrom}:{chunk_start:0,d}-{chunk_end:0,d}. ')

            # define anchor edges
            chunk_edge = max_distance
            load_beg = max(0, chunk_start - chunk_edge)
            load_end = min(chrom_size, chunk_end + chunk_edge)
            anchor_edges = np.arange(load_beg, load_end + anchor_width, anchor_width, dtype=int)
            n_anchor = len(anchor_edges) - 1
            logger.info(f'{n_anchor} anchors (width={anchor_width / 1e3:0.0f}kb) are formed over '
                        f'{chunk_chrom}:{anchor_edges[0]:0,d}-{anchor_edges[-1]:0,d}. '
                        f'This area includes {chunk_edge / 1e6:0.2f}mb "edge" anchors.')

            # prepare Triplet selector
            logger.info(f'Loading triplets from: {self.triplet_path}')
            triplets_input = triplets_obj.fetch(
                chrom_a=chunk_chrom, start_a=anchor_edges[0], end_a=anchor_edges[-1] - 1,
                chrom_b=chunk_chrom, start_b=anchor_edges[0], end_b=anchor_edges[-1] - 1,
                chrom_c=chunk_chrom, start_c=anchor_edges[0], end_c=anchor_edges[-1] - 1,
                columns=['start_A', 'start_B', 'start_C', 'count_ABC'],
                exclusive=False,
            )
        # pd.DataFrame(triplets_input)

        # map triplets to data cubes
        logger.info('Mapping triplets to a cube of counts ...')
        counts_upper = triplets_input.to_cube(anchor_edges=anchor_edges, value_column='count_ABC')
        counts_upper[np.isnan(counts_upper)] = 0
        counts_upper = counts_upper.astype(int)
        logger.info(f'In total, {counts_upper.sum():,d} triplet contacts are collected.')

        # store counts to a full symmetric cube, ignoring invalid values to prepare for smoothening
        logger.info(f'Making the cube of counts symmetric ...')
        symmetric = np.zeros([n_anchor, n_anchor, n_anchor], dtype=float)
        for ai in range(n_anchor):
            for aj in range(ai, n_anchor):
                for ak in range(aj, n_anchor):
                    if not min_distance <= anchor_edges[aj] - anchor_edges[ai] <= max_distance:
                        continue
                    if not min_distance <= anchor_edges[ak] - anchor_edges[ai] <= max_distance:
                        continue
                    if not min_distance <= anchor_edges[ak] - anchor_edges[aj] <= max_distance:
                        continue
                    symmetric[ai, aj, ak] = counts_upper[ai, aj, ak]
                    symmetric[ai, ak, aj] = counts_upper[ai, aj, ak]
                    symmetric[aj, ai, ak] = counts_upper[ai, aj, ak]
                    symmetric[aj, ak, ai] = counts_upper[ai, aj, ak]
                    symmetric[ak, ai, aj] = counts_upper[ai, aj, ak]
                    symmetric[ak, aj, ai] = counts_upper[ai, aj, ak]

        # smooth the counts
        logger.info(f'Smoothening the cube using a Gaussian kernel. Parameters: '
                    f'width={kernel_width:d}, '
                    f'scale={kernel_scale:0.2f}')
        kernel = gauss_kernel_3d(kernel_width, kernel_scale)
        # print((kernel[:, :, kernel_width // 2] * 1000).astype(int))
        smoothed = signal.convolve(symmetric, kernel, mode='same')
        # smoothed = np.round(smoothed, decimals=2)
        # temp = ndimage.convolve(cube.astype(float), kernel, mode='constant')
        # print(np.abs(smoothed - temp).sum())
        # smoothed = np.round(smoothed, 2)

        # collect valid values from smoothed cube
        logger.info(f'Collecting the upper-triangle of the smoothed counts ...')
        smoothed_triu = np.zeros([n_anchor, n_anchor, n_anchor], dtype=float)
        for ai in range(n_anchor):
            for aj in range(ai, n_anchor):
                for ak in range(aj, n_anchor):
                    # here we store everything, including smaller or larger distances:
                    # the farther or closer coverage is valid, as this coverage is from valid tripliets
                    # that are leaked/smoothed to invalid areas. So the leaked counts are also informative.
                    # if not min_distance <= anchor_edges[aj] - anchor_edges[ai] <= max_distance:
                    #     continue
                    # if not min_distance <= anchor_edges[ak] - anchor_edges[ai] <= max_distance:
                    #     continue
                    # if not min_distance <= anchor_edges[ak] - anchor_edges[aj] <= max_distance:
                    #     continue

                    smoothed_triu[ai, aj, ak] = smoothed[ai, aj, ak]
                    # in the counts below, we consider all directions, no need to store other axes
        # del smoothed

        # from matplotlib import pyplot as plt, cm
        # i = 20
        # plt.close('all')
        # cmap = cm.get_cmap('hot')
        # cmap.set_under('#0000ff')
        # clim = [0.1, np.nanpercentile(counts[i], q=100)]
        #
        # fig_h, axes = plt.subplots(1, 4, figsize=(15, 5), sharex='all', sharey='all')
        # axes = axes.ravel()
        # axes[0].imshow(counts[i, :120, :120], vmin=clim[0], vmax=clim[1], cmap=cmap)
        # axes[0].set_title(clim)
        # axes[1].imshow(symmetric[i, :120, :120], vmin=clim[0], vmax=clim[1], cmap=cmap)
        # axes[2].imshow(smoothed[i, :120, :120], vmin=clim[0], vmax=clim[1], cmap=cmap)
        # axes[3].imshow(upper_tri[i, :120, :120], vmin=clim[0], vmax=clim[1], cmap=cmap)

        # defining output triplets
        triplet_columns = [
            'start_A', 'start_B', 'start_C',
            f'{label}_A', f'{label}_B', f'{label}_C',
            f'{label}_AB', f'{label}_AC', f'{label}_BC',
            f'{label}_ABC', f'{label}_surround',
        ]
        triplets_output = Triplets(columns=triplet_columns)

        # storing Triplets counts
        logger.log_time = 0
        n_store = 0
        for ai in range(n_anchor):
            if not chunk_start <= anchor_edges[ai] < chunk_end:
                continue
            if anchor_edges[ai] > chrom_size:
                logger.debug(f'End of chromosome {chunk_chrom} ({chrom_size:,d} bp) is reached, ignoring the rest.')
                break
            for aj in range(ai, n_anchor):
                for ak in range(aj, n_anchor):
                    if not anchor_edges[aj] - anchor_edges[ai] <= max_distance:
                        continue
                    if not anchor_edges[ak] - anchor_edges[ai] <= max_distance:
                        continue
                    if not anchor_edges[ak] - anchor_edges[aj] <= max_distance:
                        continue
                    if time.time() - logger.log_time > configurations.configs['log_interval']:
                        logger.log_time = time.time()
                        logger.info(f'\tcurrently at {anchor_edges[ai]:,d} ({ai:3d}/{n_anchor}) anchor, and cached {n_store:,d} smoothed counts')

                    triplets_output['start_A'].append(anchor_edges[ai])
                    triplets_output['start_B'].append(anchor_edges[aj])
                    triplets_output['start_C'].append(anchor_edges[ak])

                    triplets_output[f'{label}_A'].append(smoothed_triu[ai, :, :].sum() + smoothed_triu[:, ai, :].sum() + smoothed_triu[:, :, ai].sum())
                    triplets_output[f'{label}_B'].append(smoothed_triu[aj, :, :].sum() + smoothed_triu[:, aj, :].sum() + smoothed_triu[:, :, aj].sum())
                    triplets_output[f'{label}_C'].append(smoothed_triu[ak, :, :].sum() + smoothed_triu[:, ak, :].sum() + smoothed_triu[:, :, ak].sum())

                    triplets_output[f'{label}_AB'].append(smoothed_triu[ai, aj, :].sum() + smoothed_triu[ai, :, aj].sum() + smoothed_triu[:, ai, aj].sum())
                    triplets_output[f'{label}_AC'].append(smoothed_triu[ai, ak, :].sum() + smoothed_triu[ai, :, ak].sum() + smoothed_triu[:, ai, ak].sum())
                    triplets_output[f'{label}_BC'].append(smoothed_triu[aj, ak, :].sum() + smoothed_triu[aj, :, ak].sum() + smoothed_triu[:, aj, ak].sum())

                    triplets_output[f'{label}_ABC'].append(smoothed_triu[ai, aj, ak])

                    triplets_output[f'{label}_surround'].append(
                        np.sum(smoothed_triu[
                               max(0, ai - 4):ai + 5,
                               max(0, aj - 4):aj + 5,
                               max(0, ak - 4):ak + 5])
                        - np.sum(smoothed_triu[
                                 max(0, ai - 2):ai + 3,
                                 max(0, aj - 2):aj + 3,
                                 max(0, ak - 2):ak + 3])
                    )

                    n_store += 1

        # rounding the counts
        for col_name in triplets_output.keys():
            if col_name.startswith(f'{label}_'):
                logger.debug(f'Rounding "{col_name}" counts to two decimals ...')
                triplets_output[col_name] = np.round(triplets_output[col_name], decimals=2)

        # storing the collected data/information
        with TripletsContainer(container_path=self.triplet_path, exclusive=True) as triplets_obj:

            # storing triplets
            triplets_obj.store(
                triplets=triplets_output,
                chrom_a=chunk_chrom,
                exclusive=False,
                verify_coords=True,
            )

            # add the new chunk to stored_chunks
            triplets_obj.add_stored_chunk([chunk_chrom, chunk_start, chunk_end], exclusive=False)

            # add transformation details
            logger.debug(f'Storing transformation details, per column:')
            cis_path = triplets_obj.experiment_path + f'/contacts/triplets/{chunk_chrom},{chunk_chrom},{chunk_chrom}/'
            for ds_name in triplets_output.keys():
                if ds_name.startswith(f'{label}_'):
                    transformation = {
                        'class': 'transformation',
                        'id': f'{ds_name}:Gaussian_smoothing:w{kernel_width},s{kernel_scale}',
                        'type': 'Gaussian_smoothing',
                        'width': kernel_width,
                        'scale': kernel_scale,
                        'source': 'count_ABC',
                    }
                    logger.debug(f'Added to "{ds_name}": {transformation}')
                    triplets_obj.store_attribute(
                        dataset_path=f'{cis_path}/{ds_name}',
                        name='transformations',
                        value=[json.dumps(transformation)],
                        exclusive=False,
                    )

            # updating attributes if needed
            # triplets_obj.retrieve_attributes(exclusive=False)
            # triplets_obj.store_attributes(exclusive=False)

        logger.info(f'{n_store:,d} triplets are updated with "{label}" transformation successfully.')



# todo: find overlaps between fragments of a read and flag them accordingly

import time
import logging
import pathlib

import h5py
import pandas as pd
import numpy as np

from triplix.core import configurations
from triplix.core.hdf5 import DatasetIterator
from triplix.core.header import TriplixHeader
from triplix.core.bam import BAMReader
from triplix.core.utilities import generate_prog_id

__version__ = '1.0.0'
COMMAND_NAME = 'triplix.concatemer'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class Concatemers(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.options = {
        #     'display.min_rows': 5
        # }

    def __repr__(self):
        return '<Concatemers>\n' + repr(pd.DataFrame(self))

    def to_dataframe(self, *args, **kwargs):
        return pd.DataFrame(self, *args, **kwargs)


class ConcatemersContainer:

    def __init__(self, file_path):
        self.file_path = pathlib.Path(file_path).expanduser()
        assert self.file_path.is_file(), f'File not found: {file_path}'
        logger.debug(f'Opening (mode=r): {self.file_path}')
        self.h5_file = h5py.File(self.file_path, mode='r')

    def __len__(self):
        return self.h5_file['/reads/name'].shape[0]

    def iter(self, return_name=False, return_length=False):

        # initializing dataset iterators
        ds_iters = dict()
        for group_name in ['reads', 'concatemers']:
            ds_iters[group_name] = dict()
            for ds_name in self.h5_file[f'/{group_name}'].keys():
                ds_iters[group_name][ds_name] = DatasetIterator(self.h5_file, f'/{group_name}/{ds_name}')

        # iterate over concatemers/reads
        n_alignments = len(ds_iters['concatemers']['read_idx'])
        n_concatemers = len(self)
        start_idx = next(ds_iters['reads']['row_index'])
        for cctm_idx in range(n_concatemers):
            end_idx = next(ds_iters['reads']['row_index'], n_alignments)
            n_frag = end_idx - start_idx

            concatemer = Concatemers()
            for ds_name in ds_iters['concatemers'].keys():
                concatemer[ds_name] = ds_iters['concatemers'][ds_name].get_next(n_frag)

            if return_name:
                concatemer['read_name'] = ds_iters['reads']['name'].get_next(1)[0].decode()
            if return_length:
                concatemer['read_length'] = ds_iters['reads']['length'].get_next(1)[0]

            yield concatemer
            start_idx = end_idx

    def __getitem__(self, pointer):
        read_grp = self.h5_file[f'/reads']
        cctm_grp = self.h5_file[f'/concatemers']
        n_read = len(read_grp['row_index'])

        # calculate read start/end
        if isinstance(pointer, slice):
            if pointer.step is not None:
                raise NotImplementedError('Step argument is not supported.')
            if pointer.start is None:
                read_start = 0
            else:
                read_start = int(pointer.start)
            if pointer.stop is None:
                read_end = n_read
            else:
                read_end = int(pointer.stop)
        else:
            read_start = int(pointer)
            read_end = read_start + 1
        read_end = min(read_end, n_read)

        # calculate row start/end
        row_start = read_grp['row_index'][read_start]
        row_end = read_grp['row_index'][read_end]

        # load fragments columns
        concatemers = Concatemers()
        for ds_name in cctm_grp.keys():
            concatemers[ds_name] = cctm_grp[ds_name][row_start:row_end]

        # assign read columns
        read_names = np.array([name.decode() for name in read_grp['name'][read_start:read_end]])
        read_lengths = read_grp['length'][read_start:read_end]
        read_idxs = np.unique(concatemers['read_idx'], return_inverse=True)[1]
        concatemers['read_name'] = read_names[read_idxs]
        concatemers['read_length'] = read_lengths[read_idxs]

        return concatemers


class ExporterHDF5:

    def __init__(
            self, bam_path, assembly, output_dir=None,
            output_name=None, assume_sorted=False,
            cell_type=None, assay_name=None, experiment_name=None,
    ):

        self.assembly = assembly
        self.cell_type = cell_type
        self.assay_name = assay_name
        self.experiment_name = experiment_name
        if self.experiment_name is None:
            self.experiment_name = bam_path.stem

        bam_path = pathlib.Path(bam_path).expanduser()
        if output_dir is None:
            output_dir = bam_path.parent
        if output_name is None:
            output_name = self.experiment_name + '.concatemers.h5'
        self.output_path = pathlib.Path(output_dir).expanduser() / output_name
        self.output_path.parent.mkdir(parents=True, exist_ok=True)
        if self.output_path.is_file():
            self.output_path.unlink()

        # prepare the input BAM file
        self.bam_reader = BAMReader(file_path=bam_path)
        if not assume_sorted and self.bam_reader.file_handle.header['HD']['SO'] != 'queryname':
            logger.error('Input BAM file should be sorted by `read name`. Please use `samtools sort -n` to sort the BAM file first.')
            raise ValueError('BAM file is not sorted.')

        self.chrom_lengths = {'unmapped': 0}
        for chr_idx, chr_name in enumerate(self.bam_reader.chroms):
            self.chrom_lengths[chr_name] = self.bam_reader.chrom_sizes[chr_idx]

        self.chrom2idx = {chr_name: idx for idx, chr_name in enumerate(self.chrom_lengths)}
        chrom_dtype = h5py.enum_dtype(self.chrom2idx, basetype='int8')

        self.strand2num = {'+': 1, '-': -1, '*': 0}
        strand_dtype = h5py.enum_dtype(self.strand2num, basetype='int8')

        self.dataset_names, self.dataset_dtypes = zip(*[
            ('read_idx', 'uint32'),
            ('chrom_num', chrom_dtype),
            ('start', 'int64'),
            ('end', 'int64'),
            ('strand', strand_dtype),
            ('map_quality', 'uint8'),
            ('flag', 'uint16'),
            ('seq_start', 'uint32'),
            ('seq_end', 'uint32'),
        ])
        self.dataset_defaults = dict(
            maxshape=(None, ),
            chunks=(configurations.configs['concatemers']['hdf5_chunk_size'], ),
            compression="gzip",
            compression_opts=configurations.configs['concatemers']['compression_level'],
            # track_order=True,
        )

    def initialize_hdf5(self):
        logger.debug(f'Opening (mode=w): {self.output_path}')
        with h5py.File(self.output_path, mode='w') as h5_file:
            cnct_grp = h5_file.create_group('/concatemers', track_order=True)
            cnct_grp.attrs['format_version'] = __version__
            cnct_grp.attrs['software_version'] = configurations.configs['version']
            cnct_grp.attrs['creation_timestamp'] = time.strftime("%Y-%m-%d %H:%M:%S")
            cnct_grp.attrs['cell_type'] = self.cell_type
            cnct_grp.attrs['assay_name'] = self.assay_name
            cnct_grp.attrs['experiment_name'] = self.experiment_name
            cnct_grp.attrs['source_file'] = str(self.bam_reader.file_path)
            cnct_grp.attrs['assembly_name'] = self.assembly
            cnct_grp.attrs['assembly_length'] = sum(self.chrom_lengths.values())
            cnct_grp.attrs['chrom_names'] = list(self.chrom_lengths.keys())
            cnct_grp.attrs['chrom_lengths'] = list(self.chrom_lengths.values())
            cnct_grp.attrs['stored_chunks'] = [f'{chr_name}:0-{chr_end}' for chr_name, chr_end in self.chrom_lengths.items() if chr_name != 'unmapped']

            # adjust headers before inclusion
            headers = TriplixHeader(self.bam_reader.file_path)
            headers.process_history.add_pg(dict(ID=COMMAND_ID, PN=COMMAND_NAME, VN=configurations.configs['version'], PS=[headers.process_history.destination_id]))
            cnct_grp.attrs['sam_headers'] = headers.process_history.to_sam_headers()

            h5_file.create_group('/reads', track_order=True)

    def store_concatemers(self):
        read_names, read_lengths, concatemers_lst = [], [], []
        for read_idx, concatemer in enumerate(self.bam_reader.get_reads()):

            # filtering and sorting alignments
            #  256, 0x100: Not a primary alignment
            #  512, 0x200: Failed platform/vendor quality checks
            # 1024, 0x400: PCR or optical duplicate
            concatemer = filter(lambda aln: aln.flag & 0x700 == 0, concatemer)
            concatemer = sorted(concatemer, key=lambda aln: aln.seq_start)

            read_names.append(concatemer[0].read_name)
            read_lengths.append(concatemer[0].read_length)

            # append the batch if full
            concatemers_lst.extend(concatemer)
            if len(read_names) >= configurations.configs['concatemers']['batch_size']:
                logger.info(f'\tstoring concatemer #{read_idx + 1:,d} ...')
                self.append_to_hdf5(read_names, read_lengths, concatemers_lst)
                read_names, read_lengths, concatemers_lst = [], [], []

        if len(concatemers_lst) > 0:
            logger.info(f'\tstoring concatemer #{concatemers_lst[-1].read_idx + 1:,d} ...')
            self.append_to_hdf5(read_names, read_lengths, concatemers_lst)
        logger.info(f'In total, stored #{concatemers_lst[-1].read_idx + 1:,d} concatemers in: {self.output_path}')

    def append_to_hdf5(self, read_names, read_lengths, concatemers):
        ctime = time.time()

        concatemers = pd.DataFrame(concatemers)
        concatemers['chrom_num'] = concatemers['chrom'].map(self.chrom2idx)
        concatemers['strand'] = concatemers['strand'].map(self.strand2num)

        logger.debug(f'Opening (mode=a): {self.output_path}')
        with h5py.File(self.output_path, mode='a') as h5_file:

            read_grp = h5_file['/reads']
            cnct_grp = h5_file['/concatemers']
            row_indices = np.unique(concatemers['read_idx'], return_index=True)[1]
            if len(read_grp.keys()) == 0:  # no data is stored yet
                read_grp.create_dataset('row_index', data=row_indices, dtype='uint32', **self.dataset_defaults)
                read_grp.create_dataset('name', data=read_names, dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
                read_grp.create_dataset('length', data=read_lengths, dtype='uint32', **self.dataset_defaults)

                for col_name, col_dtype in zip(self.dataset_names, self.dataset_dtypes):
                    cnct_grp.create_dataset(col_name, data=concatemers[col_name], dtype=col_dtype, **self.dataset_defaults)
                    if configurations.configs['debug']:
                        logger.debug(f'\tSize of "{col_name:20s}" is: {cnct_grp[col_name].nbytes / 1e6:0.2f}mb')
            else:

                # append to `reads` group
                n_cur = len(read_grp['name'])
                n_new = len(read_names)
                collection = {
                    'row_index': len(cnct_grp['start']) + row_indices,
                    'name': read_names,
                    'length': read_lengths,
                }
                for col_name in collection:
                    read_grp[col_name].resize(n_cur + n_new, axis=0)
                    read_grp[col_name][-n_new:] = collection[col_name]

                # append to `concatemers` group
                n_cur = len(cnct_grp['start'])
                n_new = len(concatemers)
                for col_name in self.dataset_names:
                    cnct_grp[col_name].resize(n_cur + n_new, axis=0)
                    cnct_grp[col_name][-n_new:] = concatemers[col_name]
        if configurations.configs['debug']:
            logger.debug(f'Stored this batch in: {time.time() - ctime:0.2f}s')




import os
import time
import pathlib
import logging
import copy
from collections import deque
from importlib import resources
import hashlib
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

import numpy as np
from sklearn.neighbors import KDTree
import joblib

from triplix.core import configurations
from triplix.core.triplets import TripletsContainer
from triplix.core.hdf5 import HDF5Container

COMMAND_NAME = 'triplix.neighbor'
logger = logging.getLogger(COMMAND_NAME)


class NeighborFinder:

    def __init__(self, triplet_path):
        self.triplet_path = pathlib.Path(triplet_path).expanduser()

    def find_neighbors(self, factors_info, target, chroms=None, output_dir=None, output_name=None, n_samples=np.inf):

        # load correction factor configs
        possible_paths = [
            pathlib.Path(factors_info).expanduser(),
            resources.files('triplix').joinpath(f'factor_configs/{factors_info}.toml')
        ]
        for file_path in possible_paths:
            if file_path.is_file():
                factors_id = file_path.stem
                with open(file_path, 'rb') as toml_file:
                    factors = tomllib.load(toml_file)['factors']
                break
        else:
            factors_id = 'unknown'
            factors = []
            for record in factors_info.split(','):
                col_name, norm_type = record.split(':')
                factors.append({'column': col_name, 'type': norm_type})
        assert len(factors) > 0, 'At least a single factor is required for KDTree construction.'

        # define output file name
        if output_dir is None:
            output_dir = self.triplet_path.parent
        if output_name is None:
            output_name = self.triplet_path.stem + f'.kdtree.{factors_id}.joblib'
        output_path = pathlib.Path(output_dir).expanduser() / output_name
        output_path.parent.mkdir(parents=True, exist_ok=True)

        # map column names to feature indices
        feature2index = {}
        for factor in factors:
            if factor['column'] not in feature2index:
                feature2index[factor['column']] = len(feature2index)
        n_feature = len(feature2index)

        # collecting Triplets across the genome
        logger.info(f'Loading triplets from: {self.triplet_path}')
        dataset = [list() for _ in range(n_feature)]
        targets = []
        with TripletsContainer(container_path=self.triplet_path, exclusive=True) as triplets_obj:
            with HDF5Container(self.triplet_path, mode='r', exclusive=False) as container:

                # define chromosomes of interest
                triplets_ds = container.h5_file[triplets_obj.experiment_path]['contacts/triplets']
                if chroms is None:
                    cis_paths = list(triplets_ds.keys())
                else:
                    cis_paths = []
                    for chrom in chroms:
                        # if chrom == 'unmapped':
                        #     continue
                        cis_paths.append(f'{chrom},{chrom},{chrom}')

                # iterate over cis-triplets and collect relevant datasets
                for cis_path in cis_paths:
                    logger.info(f'collecting cis-triplets from: {cis_path} ...')
                    cis_ds = triplets_ds[cis_path]
                    for ds_name, feat_idx in feature2index.items():
                        logger.debug(f'\tfeature idx{feat_idx}: loading "{ds_name}", {cis_ds[ds_name].shape[0]:,d} elements ...')
                        dataset[feat_idx].extend(cis_ds[ds_name][()])

                    logger.debug(f'\ttarget: loading "{target}", {cis_ds[target].shape[0]:,d} elements ...')
                    targets.extend(cis_ds[target][()])
                    if len(targets) >= n_samples:
                        logger.info(f'Reached the maximum number of samples of {n_samples:0,.0f}. No more Triplets will be loaded.')
                        break

        # converting to numpy array
        logger.debug('Converting the dataset to Numpy 2D array')
        dataset = np.array(dataset, dtype=float).T
        targets = np.array(targets, dtype=float)
        logger.info(f'Triplet dataset has {dataset.shape[0]:,d} samples and {dataset.shape[1]:d} features.')

        # limit the number of samples, if requested
        if np.isfinite(n_samples):
            n_samples = int(n_samples)
            logger.info(f'Limiting the number of samples to: {n_samples:0,.0f}')
            dataset = dataset[:n_samples, :]
            targets = targets[:n_samples]
            logger.info(f'Triplet dataset now has {dataset.shape[0]:,d} samples and {dataset.shape[1]:d} features.')

        # performing the normalization/transformation
        logger.info('Performing normalization on:')
        for fdx, factor in enumerate(factors):
            feat_name = factor['column']
            feat_idx = feature2index[feat_name]
            factors[fdx]['source_feature_idx'] = fdx
            factors[fdx]['class'] = 'normalization'

            if factor['type'] == 'zscore':
                col_mean = np.mean(dataset[:, feat_idx])
                col_std = np.std(dataset[:, feat_idx])
                logger.info(f'{feat_name:>10}: type={factor["type"]}, mean={col_mean:8.1f}; std={col_std:8.1f}')
                dataset[:, feat_idx] = (dataset[:, feat_idx] - col_mean) / col_std
                factors[fdx]['mean'] = col_mean
                factors[fdx]['std'] = col_std
                factors[fdx]['id'] = f'{feat_name}:normalization:zscore:{fdx}'
            elif factor['type'] == 'none':
                pass
            else:
                raise ValueError(f'Unknown normalization/transformation method: {factor["type"]}')

        # constructing the KD-tree
        kdtree_time = time.time()
        logger.info('Constructing KDTree using the collected samples, shape=[{:,d} x {:,d}] ...'.format(*dataset.shape))
        kdtree = KDTree(dataset, leaf_size=50, metric='euclidean')
        logger.info('Finished in {:0.1f}s'.format(time.time() - kdtree_time))

        # produce a meta-data object
        hasher = hashlib.md5()
        hasher.update('\n'.join(triplets_obj.properties['triplix_headers']).encode())
        kdtree_metadata = dict(
            creation_timestamp=time.strftime("%Y-%m-%d %H:%M:%S"),
            triplix_version=configurations.configs['version'],
            source_hash=hasher.hexdigest(),
            source_header=triplets_obj.properties['triplix_headers'],
            n_sample=dataset.shape[0],
            n_feature=dataset.shape[1],
            factors_id=factors_id,
            factors=factors,
            feature2index=feature2index,
            target_name=target,
            target_values=targets,
        )

        # storing the KDTree object
        logger.info(f'Storing the KDTree in: {output_path}')
        joblib.dump({
            'kdtree': kdtree,
            'metadata': kdtree_metadata
        }, filename=output_path, protocol=-1, compress=('zlib', 2))
        logger.info(f'KDTree is stored successfully: file size={os.stat(output_path).st_size / 1e9:0,.1f}GB')






import pathlib
import logging
import subprocess

COMMAND_NAME = 'triplix.index'
logger = logging.getLogger(COMMAND_NAME)


class TripletIndexer:
    def __init__(self, triplet_path):
        self.triplet_path = pathlib.Path(triplet_path).expanduser().resolve()
        assert self.triplet_path.is_file(), f'File not found: {self.triplet_path}'

    def index(self):
        logger.info(f'Indexing tri-alignment container: {self.triplet_path}')

        # pairix options (v0.3.7):
        #          -s INT     sequence name column [1]
        #          -d INT     second sequence name column [null]
        #          -b INT     start1 column [4]
        #          -e INT     end1 column; can be identical to '-b' [5]
        #          -u INT     start2 column [null]
        #          -v INT     end2 column; can be identical to '-u' [null or identical to the start2 specified by -u]
        index_command = (
            f"/bin/bash -c \"pairix -s1 -d4 -b2 -e3 -u5 -v6 "
            f"{self.triplet_path} "
            f"\""
        )
        logger.debug(f'Executing: {index_command}')
        subprocess.check_call(index_command, shell=True)
        logger.info(f'Index generation is finished successfully.')




from importlib import resources, metadata
import pathlib
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

configs = {}


def load_config(config_path: str = None) -> None:
    global configs

    triplix_path = pathlib.Path(__file__).parent
    # if config_path is None:
    #     config_path = triplix_path / 'config.toml'

    if config_path is None:
        with resources.files('triplix').joinpath('config.toml').open('rb') as config_file:
            configs = tomllib.load(config_file)
        configs['version'] = metadata.version('triplix')
        configs['triplix_path'] = triplix_path
        if configs['debug']:
            print('Configs are being loaded.')




import glob
import io
import time
import pathlib
import logging
import subprocess
import copy


from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.utilities import generate_prog_id, get_random_string, merge_intervals

COMMAND_NAME = 'triplix.merge'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class TripletMerger:

    def __init__(self, input_patterns, output_dir=None, output_name=None, concatenate_only=False, check_coverage=False, sort_by=None, metadata=None):
        self.concatenate_only = concatenate_only
        self.check_coverage = check_coverage
        self.sort_by = sort_by
        operation_str = 'concatenated' if self.concatenate_only else 'merged'

        # collecting of input files details
        logger.info('Collecting Triplets headers ...')
        self.input_files = []
        for pattern in input_patterns.split(','):
            pattern = str(pathlib.Path(pattern).expanduser())
            for file_path in glob.glob(pattern):
                file = {
                    'pattern': pattern,
                    'path': pathlib.Path(file_path).expanduser(),
                }
                file['header'] = TriplixHeader(file_path=file['path'])
                assert len(file['header'].sort_order) > 0, f'An input file is unsorted: {file["path"]}'
                self.input_files.append(file)
                assert file['header'].column_names == self.input_files[0]['header'].column_names, f'Column order is consistent in: \n{self.input_files[0]["path"]}\n{file["path"]}'
        assert len(self.input_files) > 1, f'Number of input Triplet files should be > 1, only {len(self.input_files)} files are given'

        # sorting input files
        logger.info('Ordering the input files according to their chunks\' chromosomal positions ...')
        self.chrom_order = {chr_name: chr_idx for chr_idx, chr_name in enumerate(self.input_files[0]['header'].chrom_lengths)}
        if self.sort_by == 'chrom':
            self.input_files = sorted(self.input_files, key=lambda inp_file: (self.chrom_order[inp_file['header'].stored_chunks[0][0]], inp_file['header'].stored_chunks[0][1], inp_file['header'].stored_chunks[0][2]))
        logger.info(f'The following {len(self.input_files):d} files will be {operation_str}:')
        for file_idx, file in enumerate(self.input_files):
            logger.info(f'\t#{file_idx + 1:d}: {file["path"]}')

        # prepare output directory and file names
        if output_dir is None:
            output_dir = self.input_files[0]['path'].parent
        if output_name is None:
            output_name = self.input_files[0]['path'].stem.replace('.sorted', '') + f'.{operation_str}' + self.input_files[0]['path'].suffix
        self.output_path = pathlib.Path(output_dir).expanduser() / output_name
        self.output_path.parent.mkdir(parents=True, exist_ok=True)
        if self.output_path.is_file():
            self.output_path.unlink()

        # use the header from the first file, and update any given meta-data attributes
        self.header = copy.deepcopy(self.input_files[0]['header'])
        for key, value in metadata.items():
            logger.debug(f'Attribute update: {key}={value}')
            setattr(self.header, key, value)

        # adjusting PG:ID headers: making sure these tags are unique
        hash2id = {}
        for input_file in self.input_files:
            prog_suffix = f'_{get_random_string(5)}'
            id_translator = {}
            for prog_name, pg_node in input_file['header'].process_history.programs.items():
                pg_hash = pg_node.hash()
                if pg_hash in hash2id:  # have we seen this PG node before?
                    pg_id = hash2id[pg_hash]
                else:
                    if pg_node['ID'] in hash2id.values():  # can we continue using this id?
                        pg_id = pg_node['ID'] + prog_suffix
                    else:
                        pg_id = pg_node['ID']
                    hash2id[pg_hash] = pg_id
                id_translator[pg_node['ID']] = pg_id

                pg_node['ID'] = id_translator[pg_node['ID']]
                if 'PP' in pg_node:
                    pg_node['PP'] = id_translator[pg_node['PP']]
                pg_node['PS'] = [id_translator[ps_id] for ps_id in pg_node['PS']]

    def merge_triplets(self, n_thread, memory, ignore_headers):
        # assert n_thread > 2, 'Number of threads must be more than 2'

        if self.concatenate_only:
            merge_command = (
                f"/bin/bash -c \"cat "
            )
        else:
            col2num = {col_name: col_idx + 1 for col_idx, col_name in enumerate(self.input_files[0]['header'].column_names)}
            separator = r"\t"
            merge_command = (
                f"/bin/bash -c \"export LC_COLLATE=C; export LANG=C; sort "
                f"-k {col2num['chrom_A']},{col2num['chrom_A']} "
                f"-k {col2num['chrom_B']},{col2num['chrom_B']} "
                f"-k {col2num['chrom_C']},{col2num['chrom_C']} "
                f"-k {col2num['start_A']},{col2num['start_A']}n "
                f"-k {col2num['start_B']},{col2num['start_B']}n "
                f"-k {col2num['start_C']},{col2num['start_C']}n "
                f"--merge "
                f"--field-separator=$'{separator}'  "
                f"--parallel={n_thread:d} "
                f"--temporary-directory=./ "
                f"--buffer-size={memory:s} "
            )

        # add input files as stdin
        for input_file in self.input_files:
            merge_command += (
                # f"<(bgzip --decompress --stdout --threads 1 {input_file['path']} | sed -n -e '/^[^#]/,$p') "
                f"<(bgzip --decompress --stdout --threads 1 {input_file['path'].resolve()} | grep -v '^#') "
            )
        merge_command += f"\""

        # prepare the output file
        # output_file = bgzip.BGZipWriter(self.output_path, 'wt', compresslevel=3)
        # output_file = gzip.open(self.output_path, 'wt', compresslevel=3)
        # pipe_template = pipes.Template()
        # pipe_template.append("bgzip --stdout --threads 1 --compress-level=3", "--")
        # output_file = pipe_template.open(str(self.output_path), "w")

        # combine table chunks and merge the adjacent ones
        stored_chunks = []
        for inp_file in self.input_files:
            stored_chunks.extend(inp_file['header'].stored_chunks)
        self.header.stored_chunks = merge_intervals(stored_chunks, chroms_order=list(self.chrom_order))

        # check if every chromosome is fully covered by chunks
        if self.check_coverage:
            seen_chroms = set()
            for chunk in self.header.stored_chunks:
                chrom_size = self.header.chrom_lengths[chunk[0]]
                assert chunk[0] not in seen_chroms, f'Duplicated chromosomes are found in chunks: {self.header.stored_chunks}'
                assert chunk[1] == 0, f'Missing the start of {chunk[0]}. The smallest chunk starts at {chunk[1]:,d}'
                assert chunk[2] >= chrom_size, f'Missing the end of {chunk[0]} (length={chrom_size:,d}). Current chunk ends at {chunk[2]:,d}'
                seen_chroms.add(chunk[0])
            logger.debug('Coverage test: All chromosomes are fully covered.')
            for chrom in self.header.chrom_lengths:
                if chrom != 'unmapped':
                    assert chrom in seen_chroms, (
                        f'Missing {chrom} in the merged file. \n'
                        f'Only seen the following chromosomes:{seen_chroms}'
                    )
            logger.debug('Coverage test: No missing chromosome is detected.')
            for chrom in seen_chroms:
                assert chrom in self.header.chrom_lengths, (
                    f'An extra chromosome is detected: {chrom}.\n'
                    f'Known chromosomes are:{list(self.header.chrom_lengths)}'
                )
            logger.debug('Coverage test: Every chromosome (stored in the merging files) is known.')

        # preparing compressor process
        compr_command = 'bgzip --stdout --threads 2 --compress-level=3'
        output_file = open(self.output_path, "w")
        compr_proc = subprocess.Popen(compr_command, stdin=subprocess.PIPE, stdout=output_file, bufsize=-1, shell=True)
        compr_input = io.TextIOWrapper(compr_proc.stdin, "utf-8")

        # update and add triplets header
        if ignore_headers:
            # todo: to be removed, when header merger is functional
            logger.info(f'Avoid merging sam headers ...')
            self.header.process_history.programs = {}
            self.header.process_history.source_ids = []
            self.header.process_history.destination_id = None
            merge_pg = dict(
                ID=COMMAND_ID,
                PN=COMMAND_NAME,
                VN=configurations.configs['version'],
                PS=[],
            )
            self.header.process_history.add_pg(merge_pg, is_destination=True)
        else:
            logger.info(f'Merging headers ...')
            headers = [input_file['header'] for input_file in self.input_files]
            self.header.merge_process_histories(headers)
            merge_pg = dict(
                ID=COMMAND_ID,
                PN=COMMAND_NAME,
                VN=configurations.configs['version'],
                PS=[prog['ID'] for prog in self.header.process_history.programs.values() if len(prog.children_ids) == 0],
            )
            self.header.process_history.add_pg(merge_pg, is_destination=True)
        self.header.update(
            creation_timestamp=time.strftime("%Y-%m-%d %H:%M:%S"),
            format_version=TriplixHeader.FORMAT_VERSION,
        )
        compr_input.write(self.header.to_string() + '\n')
        compr_input.flush()
        output_file.flush()
        logger.info(f'The merged header is added to: {self.output_path}')

        # prepare the merge process
        # stdout_wrapper = io.TextIOWrapper(output_file, "utf-8")
        # subprocess.check_call(merge_command, shell=True, stdout=stdout_wrapper)
        # print(merge_command)
        merge_proc = subprocess.Popen(merge_command, stdout=subprocess.PIPE, bufsize=-1, shell=True)
        merge_output = io.TextIOWrapper(merge_proc.stdout, "utf-8")

        # performing the merge
        log_time = 0
        n_trialgn = 0
        operation_str = 'concatenated' if self.concatenate_only else 'merged'
        logger.info(f'Storing {operation_str} tri-alignments in: {self.output_path}')
        for triplet_str in merge_output:
            n_trialgn += 1
            if time.time() - log_time > configurations.configs['log_interval']:
                logger.info(f'\tcurrently at tri-alignment #{n_trialgn:,d}')
                log_time = time.time()
            compr_input.write(triplet_str)
        compr_input.flush()
        merge_proc.communicate()
        compr_proc.communicate()
        output_file.close()
        assert merge_proc.returncode == 0 and compr_proc.returncode == 0, 'Merge process has failed.'

        logger.info(f'In total {n_trialgn:,d} tri-alignments are {operation_str} and stored in: {self.output_path}')






import time
import logging


from triplix.core import cli
from triplix.core import configurations


logging.basicConfig(
    level=logging.DEBUG if configurations.configs['debug'] else logging.INFO,
    format='%(asctime)s %(name)s [%(levelname)s]: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)


def main():

    """Main function of Triplix"""
    # ctime = time.time()
    cli_args = cli.parse_cli_arguments()
    # print(cli_args)
    # print(f'Program ended. It took {time.time() - ctime:0.4f}s')


if __name__ == "__main__":
    main()





import io
import gzip
import time
import pathlib
import logging
import subprocess

from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.utilities import generate_prog_id

COMMAND_NAME = 'triplix.sort'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class TriAlignmentSorter:
    def __init__(self, trialign_path, output_dir=None, output_name=None):
        self.trialign_path = pathlib.Path(trialign_path).expanduser()
        self.header = TriplixHeader(file_path=trialign_path)

        if output_dir is None:
            output_dir = self.trialign_path.parent
        if output_name is None:
            output_name = self.trialign_path.stem + '.sorted.tri-alignments.tsv.bgz'
        self.output_path = pathlib.Path(output_dir).expanduser() / output_name
        self.output_path.parent.mkdir(parents=True, exist_ok=True)
        if self.output_path.is_file():
            self.output_path.unlink()

    def sort_tri_alignments(self, n_thread, memory):
        col2num = {col_name: col_idx + 1 for col_idx, col_name in enumerate(self.header.column_names)}

        separator = r"\t"
        sort_command = (
            f"/bin/bash -c \"export LC_COLLATE=C; export LANG=C; sort "
            f"-k {col2num['chrom_A']},{col2num['chrom_A']} "
            f"-k {col2num['chrom_B']},{col2num['chrom_B']} "
            f"-k {col2num['chrom_C']},{col2num['chrom_C']} "
            f"-k {col2num['start_A']},{col2num['start_A']}n "
            f"-k {col2num['start_B']},{col2num['start_B']}n "
            f"-k {col2num['start_C']},{col2num['start_C']}n "
            f"--stable "
            f"--field-separator=$'{separator}'  "
            f"--parallel={n_thread:d} "
            f"--temporary-directory=./ "
            f"--buffer-size={memory:s}"
            f"\""
        )

        # ignoring header lines
        input_file = gzip.open(self.trialign_path, 'rt')
        for line in input_file:
            if line.startswith('#column names:'):
                break

        # prepare the output file
        output_file = gzip.open(self.output_path, 'wt', compresslevel=3)

        # add Tri-alignment header
        self.header.update(
            creation_timestamp=time.strftime("%Y-%m-%d %H:%M:%S"),
            format_version=TriplixHeader.FORMAT_VERSION,
            sort_order=['chrom_A', 'chrom_B', 'chrom_C', 'start_A', 'start_B', 'start_C'],
        )
        self.header.process_history.add_pg(dict(ID=COMMAND_ID, PN=COMMAND_NAME, VN=configurations.configs['version'], PS=[self.header.process_history.destination_id]))
        output_file.write(self.header.to_string() + '\n')
        output_file.flush()

        # performing the sort
        logger.info(f'Reading tri-alignments from: {self.trialign_path}')
        logger.log_time = time.time()
        with subprocess.Popen(sort_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, bufsize=-1, shell=True) as sort_process:
            stdin_wrapper = io.TextIOWrapper(sort_process.stdin, "utf-8")
            # shutil.copyfileobj(input_file, stdin_wrapper)
            n_tri_algn = 0
            for line in input_file:
                # if line_idx % 500e3 == 0:
                n_tri_algn += 1
                if time.time() - logger.log_time > configurations.configs['log_interval']:
                    logger.log_time = time.time()
                    logger.info(f'\tcurrently at tri-alignments #{n_tri_algn:,d}')
                stdin_wrapper.write(line)
            stdin_wrapper.flush()
            sort_process.stdin.close()

            logger.info(f'Sorting {n_tri_algn:,d} tri-alignments ...')

            logger.info(f'Writing sorted tri-alignments to: {self.output_path}')
            stdout_wrapper = io.TextIOWrapper(sort_process.stdout, "utf-8")
            logger.log_time = time.time()
            for line_idx, line in enumerate(stdout_wrapper):
                # if line_idx % 500e3 == 0:
                if time.time() - logger.log_time > configurations.configs['log_interval']:
                    logger.log_time = time.time()
                    logger.info(f'\tcurrently at tri-alignments #{line_idx:,d}')
                output_file.write(line)

        assert sort_process.returncode == 0, 'Sort process has failed.'

        input_file.close()
        output_file.close()




import os
# import time
import datetime
import random
import string
import socket
from typing import Dict
import logging

import numpy as np


# from triplix.core import configurations
COMMAND_NAME = 'triplix.utilities'
logger = logging.getLogger(COMMAND_NAME)


def get_random_string(length):
    # string.ascii_lowercase, string.ascii_letters, string.punctuation
    characters = string.ascii_letters + string.digits
    return ''.join(random.choice(characters) for _ in range(length))


def generate_prog_id():
    host_name = socket.gethostname()
    pid = os.getpid()
    str_time = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
    prog_id = f'{host_name}|{pid}|{str_time}'
    # get_random_string(configurations.configs['program_id_length'])
    return prog_id


def overlap(
        query_start, query_end,
        ref_start, ref_end,
        query_chrom=None, ref_chrom=None,
        ref_left_included=True,
        ref_right_included=True,
        offset=0
):

    # sanity checks
    if not isinstance(query_start, np.ndarray):
        query_start = np.array(query_start)
    if not isinstance(query_end, np.ndarray):
        query_end = np.array(query_end)
    if not isinstance(ref_start, np.ndarray):
        ref_start = np.array(ref_start)
    if not isinstance(ref_end, np.ndarray):
        ref_end = np.array(ref_end)
    assert query_start.ndim == 0 and query_end.ndim == 0, 'Query must be only one coordinate.'
    assert ref_start.ndim == 1 and ref_end.ndim == 1, 'Reference coordinates must expressed in a single column.'
    assert len(ref_start) == len(ref_end), 'Size inconsistency in reference coordinates: [start, end]'

    # check chromosome overlap, if needed
    if query_chrom is not None:
        if not isinstance(ref_chrom, np.ndarray):
            ref_chrom = np.array(ref_chrom)
        assert len(ref_start) == len(ref_chrom), 'Size inconsistency in reference coordinates: [chrom, start]'
        is_cis = query_chrom == ref_chrom
    else:
        is_cis = np.ones(len(ref_start), dtype=bool)

    # check coordinate overlaps
    if ref_left_included:
        ovl_left = query_end >= ref_start - offset
    else:
        ovl_left = query_end > ref_start - offset
    if ref_right_included:
        ovl_right = query_start <= ref_end + offset
    else:
        ovl_right = query_start < ref_end + offset

    return is_cis & ovl_left & ovl_right


def triplet_to_cube(
        triplets: Dict,
        anchor_edges: np.ndarray,
        anchor_width=None,
        data_cols=('count_ABC',),
        start_cols=('start_A', 'start_B', 'start_C'),
        end_cols=('end_A', 'end_B', 'end_C'),
        unique_col='read_id',
        symmetric=True,
):
    def update_cube_if_unique():
        voxel_id = (idx_a, idx_b, idx_c)
        if voxel_id not in seen_read_ids:
            seen_read_ids[voxel_id] = set()

        triplet_id = triplets[unique_col][trp_idx]
        if triplet_id in seen_read_ids[voxel_id]:
            return
        seen_read_ids[voxel_id].add(triplet_id)
        update_cube()

    def update_cube():
        if not (0 <= idx_a < n_anchor and 0 <= idx_b < n_anchor and 0 <= idx_c < n_anchor):
            return

        for col_name in data_cols:
            if np.isnan(cubes[col_name][idx_a, idx_b, idx_c]):
                data = triplets[col_name][trp_idx]
            else:
                data = triplets[col_name][trp_idx] + cubes[col_name][idx_a, idx_b, idx_c]

            cubes[col_name][idx_a, idx_b, idx_c] = data
            if symmetric:
                cubes[col_name][idx_a, idx_c, idx_b] = data
                cubes[col_name][idx_b, idx_a, idx_c] = data
                cubes[col_name][idx_b, idx_c, idx_a] = data
                cubes[col_name][idx_c, idx_a, idx_b] = data
                cubes[col_name][idx_c, idx_b, idx_a] = data

    if unique_col is None:
        update_func = update_cube
    else:
        seen_read_ids = {}
        update_func = update_cube_if_unique

    # find overlapping anchors
    overlap_intervals = []
    for start_cname, end_cname in zip(start_cols, end_cols):
        anchor_starts = triplets[start_cname]
        if anchor_width is None:
            anchor_ends = triplets[end_cname]
        else:
            anchor_ends = triplets[start_cname] + anchor_width
        overlap_intervals.append(np.searchsorted(anchor_edges, [anchor_starts, anchor_ends - 1], side='right') - 1)

    # check for uniqueness of bin_idxs, if needed
    if unique_col is None:
        indices_all = np.vstack(overlap_intervals).T
        indices_unq = np.unique(indices_all, axis=0)
        assert len(indices_unq) == len(indices_unq), 'Non-unique indices are found'
        del indices_all, indices_unq

    # prepare the cube
    n_anchor = len(anchor_edges) - 1
    cubes = {col: np.full([n_anchor, n_anchor, n_anchor], fill_value=np.nan) for col in data_cols}

    # fill up the cubes
    n_triplet = len(triplets[start_cols[0]])
    single_overlaps = (
            (overlap_intervals[0][0] == overlap_intervals[0][1]) &
            (overlap_intervals[1][0] == overlap_intervals[1][1]) &
            (overlap_intervals[2][0] == overlap_intervals[2][1])
    )
    for trp_idx in range(n_triplet):
        if single_overlaps[trp_idx]:
            idx_a, idx_b, idx_c = overlap_intervals[0][0][trp_idx], overlap_intervals[1][0][trp_idx], overlap_intervals[2][0][trp_idx]
            update_func()
        else:
            range_a = range(overlap_intervals[0][0][trp_idx], overlap_intervals[0][1][trp_idx] + 1)
            range_b = range(overlap_intervals[1][0][trp_idx], overlap_intervals[1][1][trp_idx] + 1)
            range_c = range(overlap_intervals[2][0][trp_idx], overlap_intervals[2][1][trp_idx] + 1)
            for idx_a in range_a:
                for idx_b in range_b:
                    for idx_c in range_c:
                        update_func()

    return cubes


def merge_intervals(intervals, chroms_order=None, check_coverage=False, verbose=False):

    # sort intervals
    if chroms_order is not None:
        chrom2idx = {chrom: idx for idx, chrom in enumerate(chroms_order)}
        intervals = sorted(intervals, key=lambda chk: (chrom2idx[chk[0]], chk[1], chk[2]))
    n_intrv = len(intervals)
    assert n_intrv > 0

    merged = [intervals[0]]
    for idx in range(1, n_intrv):
        interval = intervals[idx]

        if merged[-1][0] != interval[0]:
            merged.append(interval)
        elif merged[-1] == interval:
            if verbose:
                logger.warning(f'Identical intervals: Ignoring "{interval}"')
        elif merged[-1][2] == interval[1]:
            merged[-1][2] = interval[2]
        elif merged[-1][2] > interval[1] and merged[-1][1] < interval[2]:
            if check_coverage:
                raise ValueError(f'Overlapping intervals are found: A={merged[-1]} and B={interval}')
            merged.append(interval)
        else:
            message = f'An uncovered area is found between "{merged[-1]}" and "{interval}"'
            if check_coverage:
                raise ValueError(message)
            merged.append(interval)

    return merged





import time
import pathlib
import logging

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt, cm, patches
import h5py
import pysam

from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.concatemers import ConcatemersContainer
from triplix.core.triplets import TripletsContainer
from triplix.core.tri_alignments import TriAlignmentsContainer
from triplix.core.utilities import overlap, triplet_to_cube

COMMAND_NAME = 'triplix.stat'
logger = logging.getLogger(COMMAND_NAME)


def collect_contact_decay_stats(concatemer_path, output_dir=None, output_name=None):
    source_path = pathlib.Path(concatemer_path).expanduser()

    # prepare output file
    if output_dir is None:
        output_dir = source_path.parent
    if output_name is None:
        output_name = source_path.name.replace('.concatemers.h5', '') + '.stats.h5'
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # prepare output options
    h5_defaults = dict(
        compression="gzip",
        compression_opts=configurations.configs['concatemers']['compression_level'],
    )

    # prepare the container
    cctm_container = ConcatemersContainer(source_path)
    concatemers_iter = cctm_container.iter(return_name=True, return_length=True)

    # criteria
    chrom_names = list(cctm_container.h5_file['concatemers'].attrs['chrom_names'])
    criteria = {
        'chrom_num': list(range(len(chrom_names))),
        'mapq': list(range(0, 60 + 1, 5)) + [np.inf],
        'cctm_n_chrom': list(range(0, 12 + 1)) + [np.inf],
        'cctm_n_frag': list(range(0, 12 + 1)) + [np.inf],
        'cctm_n_cis': list(range(0, 12 + 1)) + [np.inf],
    }
    bin_edges_lin = np.hstack([np.linspace(0, 9.5e3, 20, dtype=int), np.linspace(10e3, 0.5e6, 50, dtype=int), np.inf])
    bin_edges_log = np.hstack([0, 5000, np.logspace(4, 8, 37, base=10, dtype=int), np.inf])
    counts_lin = np.zeros([len(axis_edges) for axis_name, axis_edges in criteria.items()] + [len(bin_edges_lin)], dtype=int)
    counts_log = np.zeros([len(axis_edges) for axis_name, axis_edges in criteria.items()] + [len(bin_edges_log)], dtype=int)

    # iterate over concatemers
    logger.log_time = time.time()
    n_concatemer = cctm_container.h5_file['/reads/name'].shape[0]
    # n_alignment = len(cctm_container.h5_file['/concatemers/start'])
    logger.info(f'Iterating over {n_concatemer:,d} concatemers in: {concatemer_path}')
    for cnct_idx, concatemer in enumerate(concatemers_iter):
        if time.time() - logger.log_time > configurations.configs['log_interval']:
            logger.log_time = time.time()
            logger.info(f'\t{cnct_idx:,d}/{n_concatemer:,d} concatemers are processed. ')
        n_frag = len(concatemer['read_idx'])
        n_frag_idx = min(n_frag, len(criteria['cctm_n_frag']) - 1)

        # iterate over chromosomes
        captured_chrs, captured_freq = np.unique(concatemer['chrom_num'], return_counts=True)
        n_chrom = len(captured_chrs)
        n_chrom_idx = min(n_chrom, len(criteria['cctm_n_chrom']) - 1)
        captured_chrs = captured_chrs[captured_freq > 1]

        # iterate over chromosomes with multi-way captures
        for chrom_idx, chrom_cis in enumerate(captured_chrs):
            cis_idxs = np.where(concatemer['chrom_num'] == chrom_cis)[0]
            n_cis = len(cis_idxs)
            n_cis_idx = min(n_cis, len(criteria['cctm_n_cis']) - 1)

            for cis_idx_i in range(n_cis):
                for cis_idx_j in range(cis_idx_i + 1, n_cis):
                    fi = cis_idxs[cis_idx_i]
                    fj = cis_idxs[cis_idx_j]

                    contact_length = np.abs(concatemer['start'][fi] - concatemer['start'][fj])

                    min_mapq = min(concatemer['map_quality'][fi], concatemer['map_quality'][fj])
                    mapq_idx = np.searchsorted(criteria['mapq'], min_mapq, side='right') - 1

                    counts_lin[
                        chrom_cis,
                        mapq_idx,
                        n_chrom_idx,
                        n_frag_idx,
                        n_cis_idx,
                        np.searchsorted(bin_edges_lin, contact_length, side='right') - 1,
                    ] += 1
                    counts_log[
                        chrom_cis,
                        mapq_idx,
                        n_chrom_idx,
                        n_frag_idx,
                        n_cis_idx,
                        np.searchsorted(bin_edges_log, contact_length, side='right') - 1,
                    ] += 1
        # if cnct_idx > 500e3:
        #     break

    # prepare the output path
    logger.info(f'Storing contact counts in: {output_path}')
    logger.debug(f'Opening (mode=w): {output_path}')
    with h5py.File(output_path, mode='w') as h5_file:

        ds_axes = h5_file.create_dataset('axes', data=list(criteria.keys()), dtype=h5py.string_dtype(encoding='utf-8'), **h5_defaults)
        for ax_name, ax_edges in criteria.items():
            ds_axes.attrs[ax_name] = ax_edges

        ds_lin = h5_file.create_dataset('n_contacts_linear', data=counts_lin, dtype='uint32', **h5_defaults)
        ds_lin.attrs['bins'] = bin_edges_lin

        ds_log = h5_file.create_dataset('n_contacts_logarithmic', data=counts_log, dtype='uint32', **h5_defaults)
        ds_log.attrs['bins'] = bin_edges_log

    logger.info('All contact counts are stored successfully.')





import triplix.core.configurations as _configurations

_configurations.load_config()

from triplix.core.concatemers import (
    ConcatemersContainer,
)
from triplix.core.tri_alignments import TriAlignmentsContainer
from triplix.core.triplets import TripletsContainer

__version__ = _configurations.configs['version']
__all__ = [
    'ConcatemersContainer', ]




import os
import io
import gzip
import time
import pathlib
import logging
import subprocess
import copy
from importlib import resources
import hashlib
import json

import pypairix
import numpy as np
import pandas as pd
import scipy.ndimage as ndimage
import scipy.signal as signal
from sklearn.neighbors import KDTree
import joblib

from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.triplets import TripletsContainer
from triplix.core.utilities import generate_prog_id, overlap

COMMAND_NAME = 'triplix.enrichment'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class EnrichmentEstimator:

    def __init__(self, triplet_path, kdtree_path):
        self.triplet_path = pathlib.Path(triplet_path).expanduser()
        self.kdtree_path = pathlib.Path(kdtree_path).expanduser()
        assert self.triplet_path.is_file(), f'Triplet file is not found: {self.triplet_path}'
        assert self.kdtree_path.is_file(), f'KDTree file is not found: {self.kdtree_path}'
        logger.info(f'Using KDTree: {kdtree_path}')

    def estimate_enrichments(self, chunk_chrom, chunk_start, chunk_end, n_neighbors, view_point=None):

        # load KDTree
        logger.info(f'Loading KDTree: size={os.stat(self.kdtree_path).st_size / 1e9:0.1f}GB')
        data = joblib.load(self.kdtree_path)
        kdtree = data['kdtree']
        metadata = data['metadata']
        del data
        feature2index = metadata['feature2index']
        factors_id = metadata["factors_id"]
        n_feature = len(feature2index)
        logger.info(
            f'KDTree is loaded. Details: '
            f'factors_id="{factors_id}"; '
            f'#sample={metadata["n_sample"]:,d}; '
            f'#feature={n_feature:,d}'
        )
        assert n_feature == metadata["n_feature"]

        # load triplets
        chunk_start = int(chunk_start)
        chunk_end = int(chunk_end)
        with TripletsContainer(container_path=self.triplet_path, exclusive=True) as triplets_obj:
            chrom_lengths = dict(zip(
                triplets_obj.properties['chrom_names'],
                triplets_obj.properties['chrom_lengths']
            ))
            chrom_size = chrom_lengths[chunk_chrom]
            if chunk_start > chrom_size:
                logger.info(f'The requested window ({chunk_start / 1e6:0.2f}-{chunk_end / 1e6:0.2f}mb) '
                            f'is outside of {chunk_chrom} (length={chrom_size / 1e6:0.2f}mb). '
                            f'No processing is needed.')
                return
            logger.info(f'Area of interest is: {chunk_chrom}:{chunk_start:0,d}-{chunk_end:0,d}. ')

            # load triplets from the container
            logger.info(f'Loading triplets from: {self.triplet_path}')
            triplets = triplets_obj.fetch(
                chrom_a=chunk_chrom, start_a=chunk_start, end_a=chunk_end - 1,
                columns=['start_A', 'start_B', 'start_C'] + list(feature2index.keys()) + [metadata['target_name']],
                exclusive=False,
            )
            logger.debug(f'First fetched triplet: {triplets[0]["start_A"][0]:,d}-{triplets[0]["start_B"][0]:,d}-{triplets[0]["start_C"][0]:,d}')
            logger.debug(f'Last fetched triplet: {triplets[-1]["start_A"][0]:,d}-{triplets[-1]["start_B"][0]:,d}-{triplets[-1]["start_C"][0]:,d}')
            n_sample = len(triplets)
            logger.info(f'{n_sample:,d} triplets are loaded.')

            # sub-selecting triplets, if requested
            if view_point is not None:
                view_point = int(view_point)
                vp_end = view_point + triplets_obj.properties['anchor_width']
                logger.info(f'Sub-selecting triplets to overlap with: {view_point:,d}-{vp_end:,d}')
                # vp_start, view_start, view_end = int(21.775e6), int(21.00e6), int(23.500e6)
                # vp_start, view_start, view_end = int(119.825e6), int(119.300e6), int(121.300e6)

                # filter the collected triplets to overlap with the requested interval
                trp_tmp = triplets.sort_anchors(by=view_point, force_upper=True)
                within_interval = (
                        (overlap(query_start=view_point,   query_end=vp_end - 1,   ref_start=trp_tmp['start_A'], ref_end=trp_tmp['start_A'] + triplets_obj.properties['anchor_width'] - 1)) &
                        (overlap(query_start=chunk_start, query_end=chunk_end - 1, ref_start=trp_tmp['start_B'], ref_end=trp_tmp['start_B'] + triplets_obj.properties['anchor_width'] - 1)) &
                        (overlap(query_start=chunk_start, query_end=chunk_end - 1, ref_start=trp_tmp['start_C'], ref_end=trp_tmp['start_C'] + triplets_obj.properties['anchor_width'] - 1))
                )
                del trp_tmp
                triplets = triplets[within_interval]
                n_sample = len(triplets)
                logger.info(f'{n_sample:,d} triplets are left after filtering.')

        # prepare the dataset
        dataset = np.zeros([n_sample, n_feature], dtype=float)
        for col_name, feat_idx in feature2index.items():
            dataset[:, feat_idx] = triplets[col_name]
        observed = np.array(triplets[metadata['target_name']], dtype=float)

        # normalize the loaded features
        logger.info(f'Normalizing the loaded features ...')
        for norm_idx, norm in enumerate(metadata["factors"]):
            logger.debug(f'\tStep {norm_idx + 1}, performing: {norm}')
            assert norm['class'] == 'normalization'
            col_name = norm['column']
            feat_idx = feature2index[col_name]
            if norm['type'] == 'zscore':
                dataset[:, feat_idx] = (dataset[:, feat_idx] - norm['mean']) / norm['std']
            else:
                raise ValueError(f'Unknown normalization type: {norm["type"]}')

        # add estimation columns
        estimation_columns = [
            f'{factors_id}.neighbors_distance.med',
            f'{factors_id}.neighbors_distance.std',
            f'{factors_id}.neighbors_expected.avg',
            f'{factors_id}.neighbors_expected.std',
            f'{factors_id}.enrichment',
        ]
        for col_name in estimation_columns:
            triplets[col_name] = np.full(n_sample, fill_value=np.nan)

        # estimating the enrichments
        logger.info('Estimating enrichments ...')
        n_processed = 0
        n_ignore = 0
        logger.log_time = time.time()
        for trp_idx in range(n_sample):
            if time.time() - logger.log_time > configurations.configs['log_interval']:
                logger.log_time = time.time()
                logger.info(
                    f'{trp_idx + 1:10,d}/{n_sample:,d} currently at {chunk_chrom}:'
                    f'{triplets["start_A"][trp_idx]:,d}-'
                    f'{triplets["start_B"][trp_idx]:,d}-'
                    f'{triplets["start_C"][trp_idx]:,d} | '
                    f'triplets stats: #ignored:{n_ignore:,d}'
                )

            # find neighbors, per triplet
            nei_distances, nei_indices = kdtree.query(dataset[[trp_idx]], k=n_neighbors)
            # assert nei_distances.shape == nei_indices.shape == (1, n_neighbors)
            nei_distances = nei_distances[0]
            nei_indices = nei_indices[0]
            # np.vstack([dataset[[trp_idx]], kdtree.get_arrays()[0][nei_indices]])

            # find neighbors' target values
            bkg_obs = metadata['target_values'][nei_indices].copy()

            # clip the top/bottom 1% of the distribution
            bkg_lim = np.percentile(bkg_obs, [1, 99])
            is_valid = (bkg_lim[0] < bkg_obs) & (bkg_obs < bkg_lim[1])
            bkg_obs = bkg_obs[is_valid]
            nei_distances = nei_distances[is_valid]
            # nei_indices = nei_indices[is_val]
            if len(bkg_obs) == 0:
                n_ignore += 1
                continue

            # record neighbor details
            triplets[f'{factors_id}.neighbors_distance.med'][trp_idx] = np.median(nei_distances).round(decimals=3)
            triplets[f'{factors_id}.neighbors_distance.std'][trp_idx] = np.std(nei_distances).round(decimals=3)

            # compute enrichments
            neigh_exp_avg = np.mean(bkg_obs)
            neigh_exp_std = np.std(bkg_obs)
            triplets[f'{factors_id}.neighbors_expected.avg'][trp_idx] = neigh_exp_avg.round(decimals=3)
            triplets[f'{factors_id}.neighbors_expected.std'][trp_idx] = neigh_exp_std.round(decimals=3)

            if neigh_exp_std > 0:
                enrichment = (observed[trp_idx] - neigh_exp_avg) / neigh_exp_std
            else:
                enrichment = np.nan
            triplets[f'{factors_id}.enrichment'][trp_idx] = np.round(enrichment, decimals=3)
            n_processed += 1
        logger.info(f'Cached {n_processed:,d} enrichments')

        # delete redundant columns
        for col_name in list(triplets.keys()):
            if col_name not in ['start_A', 'start_B', 'start_C'] + estimation_columns:
                logger.debug(f'Redundant column is deleted from memory: "{col_name}"')
                del triplets[col_name]

        # storing the enrichments
        logger.info(f'Storing {n_processed:,d} triplet enrichments to: {self.triplet_path}')
        with TripletsContainer(container_path=self.triplet_path, exclusive=True) as triplets_obj:

            # storing triplets
            triplets_obj.store(
                triplets=triplets,
                chrom_a=chunk_chrom,
                exclusive=False,
                verify_coords=True,
            )

            # add processed chunk
            triplets_obj.add_stored_chunk(
                [chunk_chrom, chunk_start, chunk_end],
                exclusive=False
            )

            # add transformations
            logger.debug(f'Storing transformations:')
            cis_path = triplets_obj.experiment_path + f'/contacts/triplets/{chunk_chrom},{chunk_chrom},{chunk_chrom}/'
            for ds_name in [f'{factors_id}.enrichment']:
                transformations = list()
                # for norm_info in metadata["normalizations"]:
                #     if norm_info['column'] == ds_name:
                #         transformations.append(norm_info.copy())
                transformations.append({
                    'id': f'{ds_name}:enrichment:zscore',
                    'factors_id': factors_id,
                    'class': 'enrichment',
                    'type': 'zscore',
                    'kdtree_path': str(self.kdtree_path),
                })

                triplets_obj.store_attribute(
                    dataset_path=f'{cis_path}/{ds_name}',
                    name='transformations', value=json.dumps(transformations),
                    exclusive=False,
                )
                logger.debug(f'\t"{ds_name}": {transformations}')

        logger.info(f'Enrichment estimation is finished successfully: '
                    f'#triplet:{n_processed:,d}; '
                    f'#ignored:{n_ignore:,d}; '
                    f'file size:{os.stat(self.triplet_path).st_size / 1e9:0,.1f}GB')



import time
import pathlib
import logging

import numpy as np

from triplix.core.triplets import TripletsContainer, Triplets
from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.tri_alignments import TriAlignmentsContainer
from triplix.core.utilities import generate_prog_id

COMMAND_NAME = 'triplix.bin'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class TriAlignmentBinner:
    datasets_info = {
        'dist_AB': {'dtype': 'uint32'},
        'dist_AC': {'dtype': 'uint32'},
        'dist_BC': {'dtype': 'uint32'},
        'count_A': {'dtype': 'uint32'},
        'count_B': {'dtype': 'uint32'},
        'count_C': {'dtype': 'uint32'},
        'count_AB': {'dtype': 'uint32'},
        'count_AC': {'dtype': 'uint32'},
        'count_BC': {'dtype': 'uint32'},
        'count_ABC': {'dtype': 'uint32'},
    }

    def __init__(self, input_path, anchor_width, anchor_max_distance, output_dir=None, output_name=None):
        self.input_path = pathlib.Path(input_path).expanduser()
        self.header = TriplixHeader(file_path=input_path)
        self.output_dir = output_dir
        self.output_name = output_name

        if output_dir is None:
            output_dir = self.input_path.parent
        if output_name is None:
            output_name = self.header.experiment_name + '.triplets.h5'
        self.output_path = pathlib.Path(output_dir).expanduser() / output_name
        self.output_path.parent.mkdir(parents=True, exist_ok=True)

        # open or initialize the container
        self.trpl_container = TripletsContainer(
            container_path=self.output_path,
            tri_alignments_path=self.input_path,
            anchor_width=anchor_width,
            anchor_max_distance=anchor_max_distance,
            auto_init=True,
        )

    def bin(self, chunk_chrom, chunk_start, chunk_end, mapping_quality):

        # define chunk area
        chunk_start = int(chunk_start)
        chunk_end = int(chunk_end)
        chrom_size = self.header.chrom_lengths[chunk_chrom]
        if chunk_start > chrom_size:
            logger.info(f'The requested window ({chunk_start / 1e6:0.2f}-{chunk_end / 1e6:0.2f}mb) '
                        f'is outside of {chunk_chrom} (length={chrom_size / 1e6:0.2f}mb). '
                        f'No processing is needed.')
            return
        logger.info(f'Area of interest is: '
                    f'{chunk_chrom}:{chunk_start:0,d}-{chunk_end:0,d}. ')

        # define anchors
        anchor_width = self.trpl_container.properties['anchor_width']
        anchor_max_distance = self.trpl_container.properties['anchor_max_distance']
        chunk_edge = anchor_max_distance
        load_beg = max(0, chunk_start - chunk_edge)
        load_end = min(chrom_size, chunk_end + chunk_edge)
        anchor_edges = np.arange(load_beg, load_end + anchor_width, anchor_width, dtype=int)
        n_anchor = len(anchor_edges) - 1
        logger.info(f'{n_anchor} anchors (width={anchor_width / 1e3:0.0f}kb) are formed over '
                    f'{chunk_chrom}:{anchor_edges[0]:0,d}-{anchor_edges[-1]:0,d}. '
                    f'This area includes {chunk_edge / 1e6:0.2f}mb "edge" bins.')

        # prepare Triplet selector
        logger.info(f'Loading tri-alignments from: {self.input_path}')
        tri_container = TriAlignmentsContainer(trialn_path=self.input_path)

        # loop over Tri-alignments and populate the Triplets count matrix
        logger.info('Mapping tri-alignments to triplets ...')
        counts = np.zeros([n_anchor, n_anchor, n_anchor], dtype=int)
        n_load = 0
        logger.log_time = 0
        for ai in range(n_anchor):
            if time.time() - logger.log_time > configurations.configs['log_interval']:
                logger.log_time = time.time()
                logger.info(f'\tcurrently at {anchor_edges[ai]:,d} ({ai:3d}/{n_anchor}) anchor, and loaded {n_load:,d} (unique) tri-alignments')
            if anchor_edges[ai] >= chunk_end:
                logger.debug(f'End of chunk is reached, ignoring the rest.')
                break
            if anchor_edges[ai] > chrom_size:
                logger.debug(f'End of chromosome {chunk_chrom} ({chrom_size:,d}bp) is reached, ignoring the rest.')
                break
            counts[ai, :, :] = tri_container.fetch_slice(
                vp_chrom=chunk_chrom, vp_start=anchor_edges[ai], vp_end=anchor_edges[ai + 1] - 1,
                anchor_edges=anchor_edges,
                anchor_max_distance=anchor_max_distance,
                mapping_quality=mapping_quality
            )
            n_load += np.sum(counts[ai, :, :])

        if n_load == 0:
            logger.warning('No tri-alignment was found in this interval.')
        else:
            logger.info(f'In total, loaded {n_load:0,d} tri-alignments, and used {counts.sum():,d} filtered tri-alignments.')

        # defining triplets
        triplet_columns = [
            'start_A', 'start_B', 'start_C',
            'distance_AB', 'distance_AC', 'distance_BC',
            'count_A', 'count_B', 'count_C',
            'count_AB', 'count_AC', 'count_BC',
            'count_ABC'
        ]
        triplets = Triplets(columns=triplet_columns)

        # storing triplet counts
        logger.info(f'Storing triplets to: {self.output_path}')
        logger.log_time = 0
        n_cache = 0
        n_store = 0
        batch_size = configurations.configs['triplets']['batch_size']
        for ai in range(n_anchor):
            if anchor_edges[ai] < chunk_start:
                continue
            if anchor_edges[ai] >= chunk_end:
                logger.debug(f'End of chunk is reached, ignoring the rest.')
                break
            if anchor_edges[ai] > chrom_size:  # overlap with the end of chromosome generates the last triplet.
                logger.debug(f'End of chromosome {chunk_chrom} ({chrom_size:,d} bp) is reached, ignoring the rest.')
                break
            for aj in range(ai, n_anchor):
                for ak in range(aj, n_anchor):
                    if anchor_edges[ak] - anchor_edges[ai] > anchor_max_distance:
                        break
                    if time.time() - logger.log_time > configurations.configs['log_interval']:
                        logger.log_time = time.time()
                        logger.info(f'\tcurrently at {anchor_edges[ai]:,d} ({ai:3d}/{n_anchor}) anchor, and cached {n_cache:,d} triplets')

                    triplets['start_A'].append(anchor_edges[ai])
                    triplets['start_B'].append(anchor_edges[aj])
                    triplets['start_C'].append(anchor_edges[ak])

                    triplets['distance_AB'].append(anchor_edges[aj] - anchor_edges[ai])
                    triplets['distance_AC'].append(anchor_edges[ak] - anchor_edges[ai])
                    triplets['distance_BC'].append(anchor_edges[ak] - anchor_edges[aj])

                    triplets['count_A'].append(counts[ai, :, :].sum() + counts[:, ai, :].sum() + counts[:, :, ai].sum())
                    triplets['count_B'].append(counts[aj, :, :].sum() + counts[:, aj, :].sum() + counts[:, :, aj].sum())
                    triplets['count_C'].append(counts[ak, :, :].sum() + counts[:, ak, :].sum() + counts[:, :, ak].sum())

                    triplets['count_AB'].append(counts[ai, aj, :].sum() + counts[ai, :, aj].sum() + counts[:, ai, aj].sum())
                    triplets['count_AC'].append(counts[ai, ak, :].sum() + counts[ai, :, ak].sum() + counts[:, ai, ak].sum())
                    triplets['count_BC'].append(counts[aj, ak, :].sum() + counts[aj, :, ak].sum() + counts[:, aj, ak].sum())

                    triplets['count_ABC'].append(counts[ai, aj, ak])
                    n_cache += 1

                    # store if enough are cached
                    if n_cache >= batch_size:
                        logger.info(f'\tstoring {n_cache:,d} cached triplets.')
                        self.trpl_container.store(triplets=triplets, chrom_a=chunk_chrom, verify_coords=False)
                        n_store += n_cache
                        triplets.clear()
                        n_cache = 0
        if n_cache > 0:
            logger.info(f'\tstoring {n_cache:,d} cached triplets.')
            self.trpl_container.store(triplets=triplets, chrom_a=chunk_chrom, verify_coords=False)

        # update stored chunks
        self.trpl_container.add_stored_chunk([chunk_chrom, chunk_start, chunk_end], exclusive=True)

        logger.info(f'In total, {n_store:,d} triplets are stored successfully.')

